import re
import asyncio
from typing import List, Dict, Optional, Union, Literal
from .Data import NodeCenter, NodeRect, StyleProp, BoxModel
from .domains.Runtime import RuntimeType, Script
from .exceptions import (
    CouldNotFindNodeWithGivenID, RootIDNoLongerExists, NodeNotResolved, NodeNotDescribed,
    StateError
)

def to_dict_attrs(a: list) -> Union[dict, None]:
    if not a: return None
    return {a[i]: a[i+1] for i in range(0, len(a), 2)}

class Node:
    __slots__ = (
        "page_instance", "nodeId", "parentId", "backendNodeId", "nodeType", "nodeName", "localName", "nodeValue",
        "childNodeCount", "children", "attributes", "documentURL", "baseURL", "publicId", "systemId", "internalSubset",
        "xmlVersion", "name", "value", "pseudoType", "frameId", "shadowRootType", "contentDocument", "shadowRoots",
        "templateContent", "pseudoElements", "importedDocument", "distributedNodes", "isSVG", "compatibilityMode",
        "remote_object", "isolated_id"
    )

    def __init__(
        self, page_instance, nodeId: int,
        parentId:              Optional[int] = None,
        backendNodeId:         Optional[int] = None,
        nodeType:              Optional[int] = None,
        nodeName:              Optional[str] = None,
        localName:             Optional[str] = None,
        nodeValue:             Optional[str] = None,
        childNodeCount:        Optional[int] = None,
        children:       Optional[List[dict]] = None,
        attributes:      Optional[List[str]] = None,    # Идут в списке парами ['имя атрибута', 'значение атрибута', 'имя атрибута', 'значение атрибута', ... ]
        documentURL:           Optional[str] = None,
        baseURL:               Optional[str] = None,
        publicId:              Optional[str] = None,
        systemId:              Optional[str] = None,
        internalSubset:        Optional[str] = None,
        xmlVersion:            Optional[str] = None,
        name:                  Optional[str] = None,
        value:                 Optional[str] = None,
        pseudoType:            Optional[str] = None,    # Возможные варианты: first-line, first-letter, before, after, marker, backdrop, selection, target-text, spelling-error, grammar-error, first-line-inherited, scrollbar, scrollbar-thumb, scrollbar-button, scrollbar-track, scrollbar-track-piece, scrollbar-corner, resizer, input-list-button
        frameId:               Optional[str] = None,    # доступен по дефолту в свойствах второго потомка рута  root.children[1].frameId
        shadowRootType:        Optional[str] = None,
        contentDocument:      Optional[dict] = None,
        shadowRoots:  Optional[List["Node"]] = None,    # Появляются так же у <input /> вместо 'children'
        templateContent:    Optional["Node"] = None,
        pseudoElements: Optional[List["Node"]] = None,
        importedDocument:    Optional["Node"] = None,
        distributedNodes: Optional[List[dict]] = None,  #
        isSVG:                Optional[bool] = None,    # является ли элемент SVG-элементом
        compatibilityMode: Optional[Literal["QuirksMode", "LimitedQuirksMode", "NoQuirksMode"]] = None

    ):
        self.page_instance = page_instance  # PageEx
        self.nodeId = nodeId
        self.backendNodeId = backendNodeId
        self.nodeType = nodeType
        self.nodeName = nodeName
        self.localName = localName
        self.nodeValue = nodeValue

        self.parentId = parentId
        self.publicId = publicId
        self.systemId = systemId
        self.internalSubset = internalSubset

        self.childNodeCount = childNodeCount
        self.children = self._AddChildren(children)
        self.attributes = to_dict_attrs(attributes)
        self.frameId = frameId
        self.documentURL = documentURL
        self.baseURL = baseURL
        self.xmlVersion = xmlVersion
        self.name = name
        self.value = value
        self.pseudoType = pseudoType

        self.shadowRoots = self._AddChildren(shadowRoots)
        self.shadowRootType = shadowRootType

        self.contentDocument = self._AddChild(contentDocument)
        self.templateContent = self._AddChild(templateContent)
        self.pseudoElements = self._AddChildren(pseudoElements)
        self.importedDocument = self._AddChild(importedDocument)
        self.distributedNodes = distributedNodes
        self.isSVG = isSVG
        self.compatibilityMode = compatibilityMode

        self.remote_object: Optional['RemoteObject'] = None
        self.isolated_id = None                                         # идентификатор изолированного контекста

    def __str__(self) -> str:
        return f"<Node id={self.nodeId} localName={self.localName} childNodeCount={self.childNodeCount}>"

    def _AddChildren(self, children_list: Optional[List[dict]] = None) -> List["Node"]:
        """
        Вызывается рекурсивно всякий раз, когда описание нового узла содержит
            список потомков, а так же, когда уже созданному узлу запрашиваются
            его потомки посредством метода GetChildNodes().
        :param children_list:        Список словарей, описывающих свойства потомков.
        :return:        List[Node]
        """
        if not children_list: return []
        list_nodes = []
        for child in children_list:
            list_nodes.append(Node(self.page_instance, **child))
        return list_nodes

    def _AddChild(self, child: Union[dict, "Node", None] = None) -> Optional["Node"]:
        if not child: return None
        return Node(self.page_instance, **child) if type(child) is dict else child

    async def QuerySelector(self, selector: str, ignore_root_id_exists: bool = False) -> Optional["Node"]:
        """
        Выполняет DOM-запрос, возвращая объект найденного узла, или None.
            Эквивалент  === element.querySelector()
        https://chromedevtools.github.io/devtools-protocol/tot/DOM#method-querySelector
        :param selector:                    Селектор.
        :param ignore_root_id_exists:       Игнорировать исключение при отсутствии родительского элемента.
                                                Полезно при запросах на загружающихся страницах.
        :return:        <Node>
        """
        try:
            node_id = (
                await self.page_instance.Call("DOM.querySelector", {
                    "nodeId": self.nodeId, "selector": selector
                }))["nodeId"]
        except CouldNotFindNodeWithGivenID as e:
            if match := re.search(r"nodeId\': (\d+)", str(e)):
                if match.group(1) == str(self.nodeId):
                    if ignore_root_id_exists:
                        return None
                    raise RootIDNoLongerExists
            raise
        return Node(self.page_instance, node_id) if node_id else None

    async def QuerySelectorAll(self, selector: str, ignore_root_id_exists: bool = False) -> List["Node"]:
        """
        Выполняет DOM-запрос, возвращая список объектов найденных узлов, или пустой список.
            Эквивалент  === element.querySelectorAll()
        https://chromedevtools.github.io/devtools-protocol/tot/DOM#method-querySelectorAll
        :param selector:                    Селектор.
        :param ignore_root_id_exists:       Игнорировать исключение при отсутствии родительского элемента.
                                                Полезно при запросах на загружающихся страницах.
        :return:        [ <Node>, <Node>, ... ]
        """
        nodes = []
        try:
            for node_id in (await self.page_instance.Call("DOM.querySelectorAll", {
                        "nodeId": self.nodeId, "selector": selector
                    }))["nodeIds"]:
                nodes.append(Node(self.page_instance, node_id))
        except CouldNotFindNodeWithGivenID as e:
            if match := re.search(r"nodeId\': (\d+)", str(e)):
                if match.group(1) == str(self.nodeId):
                    if ignore_root_id_exists:
                        return []
                    raise RootIDNoLongerExists
            raise
        return nodes

    async def GetChildNodes(self, depth: int = -1, pierce: bool = False) -> None:
        """
        Запрашивает событие 'DOM.setChildNodes' для собственного узла и устанавливает слушателя.
            Как только событие будет сгенерировано для текущего идентификатора узла, слушатель
            будет отменён. Список полученных потомков узла, включая текстовые будет доступен
            через его свойство 'children'.

        !ВНИМАНИЕ! Запрос потомков у <input /> не генерирует событие 'DOM.setChildNodes',
            потому как это одиночный тег, внутри которого не может быть потомков.
        :param depth:           Глубина иерархии, до которой будут получены все потомки.
                                    По умолчанию -1 == все. Чтобы задать конкретное значение,
                                    укажите любое целое число больше нуля.
        :param pierce:          Получать содержимое теневых узлов(shadowRoots, shadowDOM)?
        :return:        None
        """
        async def catch(data: dict) -> None:
            if data["parentId"] == self.nodeId:
                self.page_instance.RemoveListenerForEvent("DOM.setChildNodes", catch)
                self.children = self._AddChildren(data["nodes"])

        self.children = None
        await self.page_instance.AddListenerForEvent("DOM.setChildNodes", catch)
        await self.RequestChildNodes(depth, pierce)
        while self.children is None: await asyncio.sleep(.01)

    async def ScrollIntoView(self, rect: dict = None) -> None:
        """
        (EXPERIMENTAL)
        Прокручивает указанный прямоугольник, в котором находится узел, если он еще не виден. После применения
            идентификатор узла становится недействителен и его придётся запрашивать вновь.
        https://chromedevtools.github.io/devtools-protocol/tot/DOM#method-scrollIntoViewIfNeeded
        :param rect:        (optional) Прямоугольник, который будет прокручиваться в поле зрения относительно
                                поля границы узла, в пикселях CSS. Если не указан, будет использоваться
                                центр узла, аналогично Element.scrollIntoView.
                                Ожидается словарь, вида: {"x": 100, "y": 100, "width", 200, "height": 200}
        :return:        None
        """
        args = {"nodeId": self.nodeId}
        if rect:
            args.update({"rect": rect})
        await self.page_instance.Call("DOM.scrollIntoViewIfNeeded", args)

    async def FocusNode(self) -> bool:
        """
        Фокусируется на элементе.
        https://chromedevtools.github.io/devtools-protocol/tot/DOM#method-focus
        :return:            Была ли фокусировка успешной
        """
        try:
            await self.page_instance.Call("DOM.focus", {"nodeId": self.nodeId})
            return True
        except:
            return False

    async def GetAttributes(self) -> Dict[str, str]:
        """
        Возвращает список атрибутов элемента.
        https://chromedevtools.github.io/devtools-protocol/tot/DOM#method-getAttributes
        :return:            {'id': 'element-id', 'style': 'position:absolute;...', ...}
        """
        return to_dict_attrs((await self.page_instance.Call("DOM.getAttributes", {"nodeId": self.nodeId}))["attributes"])

    async def RemoveAttribute(self, name: str) -> None:
        """
        Удаляет атрибут по 'name'.
        https://chromedevtools.github.io/devtools-protocol/tot/DOM#method-removeAttribute
        :param name:        Имя удаляемого атрибута.
        :return:
        """
        await self.page_instance.Call("DOM.removeAttribute", {"nodeId": self.nodeId, "name": name})

    async def RemoveNode(self) -> None:
        """
        Удаляет себя из документа.
        https://chromedevtools.github.io/devtools-protocol/tot/DOM#method-removeNode
        :return:
        """
        await self.page_instance.Call("DOM.removeNode", {"nodeId": self.nodeId})

    async def GetBoxModel(self) -> BoxModel:
        """
        Возвращает 'box-model' элемента.
        https://chromedevtools.github.io/devtools-protocol/tot/DOM#method-getBoxModel
        :return:            <BoxModel>
        """
        return BoxModel(**(await self.page_instance.Call("DOM.getBoxModel", {"nodeId": self.nodeId}))["model"])

    async def GetOuterHTML(self) -> str:
        """
        Возвращает HTML-разметку элемента включая внешние границы тега.
        https://chromedevtools.github.io/devtools-protocol/tot/DOM#method-getOuterHTML
        :return:            Строка HTML-разметки. '<div>Inner div text</div>'.
        """
        return (await self.page_instance.Call("DOM.getOuterHTML", {"nodeId": self.nodeId}))["outerHTML"]

    async def GetInnerHTML(self) -> str:
        """
        Возвращает HTML-разметку элемента НЕ включая внешние границы тега.
        :return:            Строка HTML-разметки. 'Inner div <div>text</div> with HTML'.
        """
        html = await self.GetOuterHTML()
        if m := re.match(r"^<.*?>(.*)<.*?>$", html):
            return m.group(1)
        return html

    async def GetInnerText(self, with_new_line_symbols: bool = True) -> str:
        """
        Возвращает текстовое содержимое элемента.
        :param with_new_line_symbols:     (optional) — Оставлять символы новой строки?
        :return:                        Текст.
        """
        pattern = r"<.*?>" if with_new_line_symbols else r"(<.*?>)|\n"
        html = await self.GetOuterHTML()
        return re.sub(pattern, "", html).strip()

    async def SetOuterHTML(self, outerHTML: str) -> None:
        """
        Устанавливает HTML-разметку для элемента. !Внимание! После этого преобразования, элемент
            получит новый идентификатор и повтороно обратиться к нему уже не получится.
        https://chromedevtools.github.io/devtools-protocol/tot/DOM#method-setOuterHTML
        :param outerHTML:       Строка HTML-разметки. '<div>Inner div text</div>'.
        :return:
        """
        await self.page_instance.Call("DOM.setOuterHTML", {"nodeId": self.nodeId, "outerHTML": outerHTML})

    async def MoveTo(
            self, targetNodeId: int,
            insertBeforeNodeId: Optional[int] = None
    ) -> None:
        """
        Перемещает узел в новый контейнер, помещает его перед заданным якорем. В результате,
            внутренний идентификатор узла будет сменён.
        https://chromedevtools.github.io/devtools-protocol/tot/DOM#method-moveTo
        :param targetNodeId:        Идентификатор элемента, куда будет помещён элемент.
        :param insertBeforeNodeId   (optional) Попытается поместить в этот элемент, но если он
                                        не будет найден, то перемещаемый элемент становится
                                        последним дочерним элементом 'targetNodeId'.
        :return:
        """
        args = {"nodeId": self.nodeId, "targetNodeId": targetNodeId}
        if insertBeforeNodeId:
            args.update({"insertBeforeNodeId": insertBeforeNodeId})
        self.nodeId = (await self.page_instance.Call("DOM.moveTo", args))["nodeId"]

    async def CopyTo(
            self, targetNodeId: int, insertBeforeNodeId: Optional[int] = None
    ) -> "Node":
        """
        Создает глубокую копию текущего узла и помещает ее в 'targetNodeId' перед 'insertBeforeNodeId',
            если последний указан и будет найден. Возвращает склонированный узел.
        https://chromedevtools.github.io/devtools-protocol/tot/DOM#method-moveTo
        :param targetNodeId:        Идентификатор элемента, куда будет скопирован текущий элемент.
        :param insertBeforeNodeId   (optional) Попытается поместить в этот элемент, но если он
                                        не будет найден, то копируемый элемент становится
                                        последним дочерним элементом 'targetNodeId'.
        :return:                    <Node> - клона.
        """
        args = {"nodeId": self.nodeId, "targetNodeId": targetNodeId}
        if insertBeforeNodeId:
            args.update({"insertBeforeNodeId": insertBeforeNodeId})
        node_id = (await self.page_instance.Call("DOM.copyTo", args))["nodeId"]
        return Node(node_id, self.page_instance)

    async def GetContentQuads(
            self,
            backendNodeId: Optional[int] = None,
                 objectId: Optional[str] = None
    ) -> List[List[int]]:
        """
        (EXPERIMENTAL)
        Возвращает квадраты, которые описывают положение узла на странице. Этот метод может вернуть
            несколько квадратов для встроенных узлов.
        https://chromedevtools.github.io/devtools-protocol/tot/DOM/#method-getContentQuads
        :param backendNodeId:       Бэкенд идентификатор нода.
        :param objectId:            Идентификатор объекта JavaScript обертки узла.
        :return:                    quads - Массив четырехугольных вершин, где за  x всегда следует y,
                                        указывая точки по часовой стрелке.
        """
        args = {"nodeId": self.nodeId}
        if not args and backendNodeId is not None:
            args.update({"backendNodeId": backendNodeId})
        if not args and objectId is not None:
            args.update({"objectId": objectId})
        if args:
            return (await self.page_instance.Call("DOM.getContentQuads", args))["quads"]

    async def SetAttributeValue(self, attributeName: str, value: str) -> None:
        """
        Устанавливает атрибут для элемента с данным идентификатором.
            Например: await node.SetAttributeValue('class', 'class-name')
        https://chromedevtools.github.io/devtools-protocol/tot/DOM#method-setAttributeValue
        :param attributeName:   Attribute name.
        :param value:           Attribute value.
        :return:
        """
        await self.page_instance.Call("DOM.setAttributeValue", {"nodeId": self.nodeId, "name": attributeName, "value": value})

    async def RequestChildNodes(self, depth: int = 1, pierce: bool = False) -> None:
        """
        Запрашивает, чтобы дочерние элементы узла с данным идентификатором возвращались
            вызывающей стороне в форме событий DOM.setChildNodes, при которых извлекаются
            не только непосредственные дочерние элементы, но и все дочерние элементы до
            указанной глубины.
        https://chromedevtools.github.io/devtools-protocol/tot/DOM#method-requestChildNodes
        :param depth:           (optional) - Максимальная глубина, на которой должны быть получены
                                    дочерние элементы, по умолчанию равна 1. Используйте -1 для
                                    всего поддерева или укажите целое число больше 0.
        :param pierce:           (optional) - Должны ли проходить фреймы и теневые корни при
                                    возврате поддерева (по умолчанию false).
        :return:
        """
        args = {"nodeId": self.nodeId, "depth": depth, "pierce": pierce}
        await self.page_instance.Call("DOM.requestChildNodes", args)

    async def SetAttributesAsText(self, text: str, name: str = "") -> None:
        """
        Устанавливает атрибуты для элемента с заданным идентификатором. Этот метод полезен, когда пользователь
            редактирует некоторые существующие значения и типы атрибутов в нескольких парах имя/значение атрибута.
            Например(!не проверено!): await node.SetAttributesAsText("'class: class-name' 'data-id: new data-id'")
        https://chromedevtools.github.io/devtools-protocol/tot/DOM#method-setAttributesAsText
        :param text:            Текст с рядом атрибутов. Будет анализировать этот текст с помощью HTML-парсера.
        :param name:            (optional) Имя атрибута для замены новыми атрибутами, полученными из текста,
                                    в случае успешного анализа текста.
        :return:
        """
        args = {"nodeId": self.nodeId, "text": text}
        if name:
            args.update({"name": name})
        await self.page_instance.Call("DOM.setAttributesAsText", args)

    async def SetFileInputFiles(self, files: List[str]) -> None:
        """
        Устанавливает файлы для '<input />'- элемента с заданным идентификатором.
        https://chromedevtools.github.io/devtools-protocol/tot/DOM#method-setFileInputFiles
        :param files:           Список файлов. [ "path", "path", ... ]
        :return:
        """
        await self.page_instance.Call("DOM.setFileInputFiles", {"nodeId": self.nodeId, "files": files})

    async def SetNewName(self, name: str) -> None:
        """
        Устанавливает новое имя узла для узла. В результате узел получит новый иденификатор.
            Например: await node.SetNewName('span')
        https://chromedevtools.github.io/devtools-protocol/tot/DOM#method-setNodeName
        :param name:            Новое имя элемента(узла).
        :return:
        """
        self.nodeId = (await self.page_instance.Call("DOM.setNodeName", {"nodeId": self.nodeId, "name": name}))["nodeId"]

    async def SetNewValue(self, value: str) -> None:
        """
        Устанавливает значение.
        https://chromedevtools.github.io/devtools-protocol/tot/DOM#method-setNodeValue
        :param value:           Новое значение элемента(узла).
        :return:
        """
        await self.page_instance.Call("DOM.setNodeValue", {"nodeId": self.nodeId, "value": value})

    async def SetInspectedNode(self) -> None:
        """
        Позволяет консоли обращаться к этому узлу с через $ x (см. Более подробную
            информацию о функциях $ x в API командной строки).
        https://chromedevtools.github.io/devtools-protocol/tot/DOM#method-setInspectedNode
        :return:        None
        """
        await self.page_instance.Call("DOM.setInspectedNode", {"nodeId": self.nodeId})

    async def CollectClassNames(self) -> List[str]:
        """
        Собирает имена классов для выбранного узла и всех его потомков.
        https://chromedevtools.github.io/devtools-protocol/tot/DOM/#method-collectClassNamesFromSubtree
        :return:                Список имён классов.
        """
        return (await self.page_instance.Call("DOM.collectClassNamesFromSubtree", {"nodeId": self.nodeId}))["classNames"]

    async def GetNodesForSubtreeByStyle(
            self, computedStyles: List[dict],
            pierce: bool = False
    ) -> List["Node"]:
        """
        Находит узлы с заданным вычисленным стилем в поддереве текущего узла.
        https://chromedevtools.github.io/devtools-protocol/tot/DOM/#method-getNodesForSubtreeByStyle
        :param computedStyles:  Список вычисляемых CSS-свойств, соответствие которым будет проверяться.
                                    Например, найти все узлы, ширина которых = 50px, а высота = 15px
                                    [{'name': 'width', 'value': '50px'}, {'name': 'height', 'value': '15px'}]
        :param pierce:          (optional) - Следует ли исследовать так же фреймы и shadow-DOM.
        :return:                Список узлов.
        """
        args = {"nodeId": self.nodeId, "computedStyles": computedStyles}
        if pierce: args.update({"pierce": True})
        nodes = []
        for node_id in (await self.page_instance.Call("DOM.getNodesForSubtreeByStyle", args))["nodeIds"]:
            nodes.append(Node(node_id, self.page_instance))
        return nodes

    async def GetComputedStyle(self) -> List[StyleProp]:
        """
        Возвращает список <NodeProp> описывающих свойства указанного Узла.
        https://chromedevtools.github.io/devtools-protocol/tot/CSS/#method-getComputedStyleForNode
        :return:            [ <StyleProp>, ... ]
        """
        r = (await self.page_instance.Call("CSS.getComputedStyleForNode", {"nodeId": self.nodeId}))["computedStyle"]
        return [StyleProp(**i) for i in r]

    async def GetInlineStyles(self) -> Dict[str, dict]:
        """
        Возвращает стили, определенные встроенными (явно в атрибуте style и неявно, используя атрибуты DOM) для
            узла DOM, идентифицированного с помощью nodeId.
            !ВНИМАНИЕ! Требует включения уведомлений домена CSS и DOM.
        https://chromedevtools.github.io/devtools-protocol/tot/CSS/#method-getInlineStylesForNode
        :return:            { inlineStyle: {}, attributeStyle: {} }
        """
        return await self.page_instance.Call("CSS.getInlineStylesForNode", {"nodeId": self.nodeId})

    # ==================================================================================================================

    async def GetCenter(self) -> NodeCenter:
        """ Возвращает координаты центра узла """
        quad = (await self.GetContentQuads())[0]
        x = (quad[2] - quad[0]) // 2 + quad[0]
        y = (quad[7] - quad[1]) // 2 + quad[1]
        return NodeCenter(x, y)

    async def GetRect(self) -> NodeRect:
        """ Возвращает <NodeRect> свойств, описывающих пространственное положение узла """
        q = (await self.GetContentQuads())[0]
        return NodeRect(q[0], q[1], q[2] - q[0], q[7] - q[1], q[0], q[2], q[1], q[7])

    async def Click(self, delay: float = None, has_link: bool = False) -> None:
        """ Кликает в середину себя """
        if has_link:
            self.page_instance.loading_state = "do_navigate"
        center = await self.GetCenter()
        await self.page_instance.action.MouseMoveTo(center.x, center.y)
        await self.page_instance.action.ClickTo(center.x, center.y, delay)

    async def Describe(self, depth: Optional[int] = None, pierce: Optional[bool] = None) -> None:
        """
        Переописывает себя получая больше подробностей. Не требует включения домена. Не начинает
            отслеживать какие-либо объекты, можно использовать для автоматизации.
        https://chromedevtools.github.io/devtools-protocol/tot/DOM/#method-describeNode
        :param depth:               Максимальная глубина, на которой должны быть извлечены
                                        дочерние элементы, по умолчанию равна 1. Используйте
                                        -1 для всего поддерева или укажите целое число больше 0.
        :param pierce:              Должны ли проходиться iframes и теневые корни при возврате
                                        поддерева (по умолчанию false).
        :return:            <Node>
        """
        args = dict(nodeId=self.nodeId)
        if depth is not None: args.update(depth=depth)
        if pierce is not None: args.update(pierce=pierce)
        result = await self.page_instance.Call("DOM.describeNode", args)
        # print("DOM.describeNode", result)

        for k, v in result["node"].items():
            if k in ["children", "shadowRoots", "pseudoElements"]:
                setattr(self, k, self._AddChildren(v))
            elif k in ["contentDocument", "templateContent", "importedDocument"]:
                setattr(self, k, self._AddChild(v))
            else:
                setattr(self, k, v)

    async def Resolve(self) -> None:
        """ Получает ссылку на объект JavaScript для ноды. """
        if self.backendNodeId is None:
            raise NodeNotDescribed
        args = dict(backendNodeId=self.contentDocument.backendNodeId)
        result: dict = await self.page_instance.Call("DOM.resolveNode", args)
        # print("DOM.resolveNode", result)
        self.remote_object = RuntimeType.RemoteObject(**result.get("object"))

    async def Request(self) -> "Node":
        """ Запрашивает ноду по ссылке на её оригинальный JavaScript объект. """
        if self.remote_object is None:
            raise NodeNotResolved
        args = dict(objectId=self.remote_object.objectId)
        result: dict = await self.page_instance.Call("DOM.requestNode", args)
        # print(result)
        return Node(
            self.page_instance, result.get("nodeId"),
            frameId=self.frameId,
            nodeName=self.nodeName,
            localName=self.localName,
            attributes=self.attributes,
            contentDocument=self.contentDocument
        )

    async def RequestMirror(self) -> "Node":
        """
        Создаёт JavaScript-объект для этой ноды и запрашивает ноду по ссылке на этот объект.
            Таким образом можно получать ноды для изолированных DOM-элементов, вроде iframe.
            Например:
                frame_node = await page.QuerySelector("iframe")     # находим iframe
                resolved_node = await frame_node.RequestMirror()    # получаем доступ к нему
                needle_node = resolved_node.QuerySelector("#needle-selector-in-iframe")
        """
        await self.Describe()
        await self.Resolve()
        return await self.Request()

    async def BuildScript(self, expression: str) -> "Script":
        if not self.page_instance.runtime_enabled:
            raise StateError("Domain 'Runtime' — must be enabled. Like: "
                             "await instance.RuntimeEnable(True)")
        if not self.page_instance.context_manager.is_watch:
            raise StateError("Watch for execution contexts must be enabled with domain 'Runtime'. Like: "
                             "await instance.RuntimeEnable(True)")
        if not self.frameId:
            await self.Describe()
        if not (context := self.page_instance.context_manager.GetDefaultContext(
                self.frameId or self.page_instance.page_id)):
            raise StateError("Something went wrong, or context was not obtained on creation.")

        return Script(self.page_instance, expression, context)





try:
    import ujson as json
except ModuleNotFoundError:
    import json
from abc import ABC, abstractmethod
from typing import Optional, Union, List, Literal, Callable
from dataclasses import dataclass, field
from ..Data import DomainEvent
from ..exceptions import PromiseEvaluateError, highlight_promise_error


class Runtime(ABC):
    """
    #   https://chromedevtools.github.io/devtools-protocol/tot/Runtime
    """
    __slots__ = ()

    def __init__(self):
        self.runtime_enabled = False
        self.context_manager = ContextManager()

    @property
    def connected(self) -> bool:
        return False

    @property
    def verbose(self) -> bool:
        return False

    @property
    def page_id(self) -> str:
        return ""

    async def GetProperties(
            self, objectId: str,
            skip_complex_types: bool = True,
            ownProperties: Optional[bool] = None,
            accessorPropertiesOnly: Optional[bool] = None,
            generatePreview: Optional[bool] = None,
            nonIndexedPropertiesOnly: Optional[bool] = None,
    ) -> List["RuntimeType.PropertyDescriptor"]:
        """ Возвращает свойства заданного объекта. Группа объектов результата наследуется
        от целевого объекта.
        https://chromedevtools.github.io/devtools-protocol/tot/Runtime/#method-getProperties
        :param objectId:                    Идентификатор объекта, для которого возвращаются свойства.
        :param ownProperties:               Если true, возвращает свойства, принадлежащие только самому
                                                элементу, а не его цепочке прототипов.
        :param accessorPropertiesOnly:      Если true, возвращает только свойства доступа (с геттером/сеттером);
                                                внутренние свойства также не возвращаются.
        :param generatePreview:             Должен ли быть создан предварительный просмотр для результатов.
        :param nonIndexedPropertiesOnly:    Если true, возвращает только неиндексированные свойства.
        :return:    {
            "result": array[ PropertyDescriptor ],
            "internalProperties":  list[ InternalPropertyDescriptor ],
            "privateProperties":  list[ PrivatePropertyDescriptor ],
            "exceptionDetails": dict{ ExceptionDetails }
        }
        """
        args = {"objectId": objectId}
        if ownProperties: args.update({"ownProperties": ownProperties})
        if accessorPropertiesOnly: args.update({"accessorPropertiesOnly": accessorPropertiesOnly})
        if generatePreview: args.update({"generatePreview": generatePreview})
        if nonIndexedPropertiesOnly: args.update({"nonIndexedPropertiesOnly": nonIndexedPropertiesOnly})
        response = await self.Call("Runtime.getProperties", args)
        if "exceptionDetails" in response:
            raise PromiseEvaluateError(
                highlight_promise_error(response["result"]["description"]) +
                "\n" + json.dumps(response["exceptionDetails"])
            )
        if not skip_complex_types:
            return [RuntimeType.PropertyDescriptor(**p) for p in response["result"]]

        result = []
        for p in response["result"]:
            if (subtype := p["value"].get("type")) and subtype == "function":
                continue
            result.append(RuntimeType.PropertyDescriptor(**p))
        return result


    async def AwaitPromise(
            self, promiseObjectId: str, returnByValue: bool = False, generatePreview: bool = False
    ) -> dict:
        """
        Добавляет обработчик к промису с переданным идентификатором.
        https://chromedevtools.github.io/devtools-protocol/tot/Runtime#method-awaitPromise
        :param promiseObjectId:     Идентификатор промиса.
        :param returnByValue:       (optional) Ожидается ли результат в виде объекта JSON,
                                        который должен быть отправлен по значению.
        :param generatePreview:     (optional) Должен ли предварительный просмотр
                                        генерироваться для результата.
        :return:                    {
                                        "result": dict(https://chromedevtools.github.io/devtools-protocol/tot/Runtime#type-RemoteObject)
                                        "exceptionDetails": dict(https://chromedevtools.github.io/devtools-protocol/tot/Runtime#type-ExceptionDetails)
                                    }
        """
        args = {"promiseObjectId": promiseObjectId, "returnByValue": returnByValue, "generatePreview": generatePreview}
        response = await self.Call("Runtime.awaitPromise", args)
        if "exceptionDetails" in response:
            raise PromiseEvaluateError(
                highlight_promise_error(response["result"]["description"]) +
                "\n" + json.dumps(response["exceptionDetails"])
            )
        return response["result"]

    async def CallFunctionOn(
            self, functionDeclaration: str,
            objectId: Optional[str] = None,
            arguments: Optional[list] = None,
            silent: Optional[bool] = None,
            returnByValue: Optional[bool] = None,
            generatePreview: Optional[bool] = None,
            userGesture: Optional[bool] = None,
            awaitPromise: Optional[bool] = None,
            executionContextId: Optional[int] = None,
            objectGroup: Optional[str] = None
    ) -> dict:
        """
        Вызывает функцию с заданным объявлением для данного объекта. Группа объектов результата
            наследуется от целевого объекта.
        https://chromedevtools.github.io/devtools-protocol/tot/Runtime#method-callFunctionOn
        :param functionDeclaration:     Объявление функции для вызова.
        :param objectId:                (optional) Идентификатор объекта для вызова функции.
                                            Должен быть указан либо objectId, либо executeContextId.
        :param arguments:               (optional) Аргументы. Все аргументы вызова должны
                                            принадлежать тому же миру JavaScript, что и целевой
                                            объект.
        :param silent:                  (optional) В тихом режиме исключения, выданные во время оценки,
                                            не регистрируются и не приостанавливают выполнение.
                                            Переопределяет 'setPauseOnException' состояние.
        :param returnByValue:           (optional) Ожидается ли результат в виде объекта JSON,
                                            который должен быть отправлен по значению.
        :param generatePreview:         (optional, EXPERIMENTAL) Должен ли предварительный
                                            просмотр генерироваться для результата.
        :param userGesture:             (optional) Должно ли выполнение рассматриваться как
                                            инициированное пользователем в пользовательском интерфейсе.
        :param awaitPromise:            (optional) Решено ли выполнение await для полученного значения
                                            и возврата после ожидаемого обещания.
        :param executionContextId:      (optional) Определяет контекст выполнения, в котором будет
                                            использоваться глобальный объект для вызова функции.
                                            Должен быть указан либо executeContextId, либо objectId.
        :param objectGroup:             (optional) Символическое имя группы, которое можно
                                            использовать для освобождения нескольких объектов. Если
                                            objectGroup не указан, а objectId равен, objectGroup
                                            будет унаследован от объекта.
        :return:                        { ... } - https://chromedevtools.github.io/devtools-protocol/tot/Runtime/#type-RemoteObject
        """
        args = {"functionDeclaration": functionDeclaration}
        if objectId is not None:
            args.update({"objectId": objectId})
        if arguments is not None:
            args.update({"arguments": arguments})
        if silent is not None:
            args.update({"silent": silent})
        if returnByValue is not None:
            args.update({"returnByValue": returnByValue})
        if generatePreview is not None:
            args.update({"generatePreview": generatePreview})
        if userGesture is not None:
            args.update({"userGesture": userGesture})
        if awaitPromise is not None:
            args.update({"awaitPromise": awaitPromise})
        if executionContextId is not None:
            args.update({"executionContextId": executionContextId})
        if objectGroup is not None:
            args.update({"objectGroup": objectGroup})
        response = await self.Call("Runtime.callFunctionOn", args)
        if "exceptionDetails" in response:
            raise Exception(response["result"]["description"] + "\n" + json.dumps(response["exceptionDetails"]))
        return response["result"]

    async def RuntimeEnable(self, watch_for_execution_contexts: bool = False) -> None:
        """
        Включает создание отчетов о создании контекстов выполнения с помощью события executeContextCreated.
            При включении, событие будет отправлено немедленно для каждого существующего контекста выполнения.

        Позволяет так же организовать обратную связь со страницей, посылая из её контекста, данные, в консоль.
            В этом случае будет генерироваться событие 'Runtime.consoleAPICalled':
            https://chromedevtools.github.io/devtools-protocol/tot/Runtime#event-consoleAPICalled
            {
                'method': 'Runtime.consoleAPICalled',
                'params': {
                    'type': 'log',
                    'args': [{'type': 'string', 'value': 'you console data passed was be here'}],
                    'executionContextId': 2,
                    'timestamp': 1583582949679.153,
                    'stackTrace': {
                        'callFrames': [{'functionName': '', 'scriptId': '48', 'url': '', 'lineNumber': 0, 'columnNumber': 8}]
                    }
                }
            }

        https://chromedevtools.github.io/devtools-protocol/tot/Runtime#method-enable
        :param watch_for_execution_contexts:    Регистрирует слушателей, ожидающих события создания/уничтожения
                                                    контекстов, которые можно запрашивать через
                                                    page_instance.context_manager.GetDefaultContext(frameId: str).
                                                    Должен быть включён ПЕРЕД переходом на целевой адрес.
        :return:
        """
        if not self.runtime_enabled:
            await self.Call("Runtime.enable")
            self.runtime_enabled = True

        if watch_for_execution_contexts and not self.context_manager.is_watch:
            await self.AddListenerForEvent(RuntimeEvent.executionContextCreated, self.context_manager.on_create)
            await self.AddListenerForEvent(RuntimeEvent.executionContextsCleared, self.context_manager.on_clear)
            await self.AddListenerForEvent(RuntimeEvent.executionContextDestroyed, self.context_manager.on_destroy)
            self.context_manager.is_watch = True

    async def RuntimeDisable(self) -> None:
        """
        Отключает создание отчетов о создании контекстов выполнения.
        https://chromedevtools.github.io/devtools-protocol/tot/Runtime#method-disable
        :return:
        """
        if self.runtime_enabled:
            await self.Call("Runtime.disable")
            self.runtime_enabled = False

        if self.context_manager.is_watch:
            self.RemoveListenerForEvent(RuntimeEvent.executionContextCreated, self.context_manager.on_create)
            self.RemoveListenerForEvent(RuntimeEvent.executionContextsCleared, self.context_manager.on_clear)
            self.RemoveListenerForEvent(RuntimeEvent.executionContextDestroyed, self.context_manager.on_destroy)
            self.context_manager.is_watch = False

    async def DiscardConsoleEntries(self) -> None:
        """
        Отбрасывает собранные исключения и вызовы API консоли.
        https://chromedevtools.github.io/devtools-protocol/tot/Runtime#method-discardConsoleEntries
        :return:
        """
        await self.Call("Runtime.discardConsoleEntries")

    async def ReleaseObjectGroup(self, objectGroup: str) -> None:
        """
        Освобождает все удаленные объекты, принадлежащие данной группе.
        https://chromedevtools.github.io/devtools-protocol/tot/Runtime#method-releaseObjectGroup
        :param objectGroup:             Символическое имя группы.
        :return:
        """
        await self.Call("Runtime.releaseObjectGroup", {"objectGroup": objectGroup})

    async def CompileScript(
            self, expression: str,
            sourceURL: str = "",
            persistScript: bool = True,
            executionContextId: Optional[int] = None
    ) -> str:
        """
        Компилирует выражение.
        https://chromedevtools.github.io/devtools-protocol/tot/Runtime#method-compileScript
        :param expression:              Выражение для компиляции.
        :param sourceURL:               Исходный URL для скрипта.
        :param persistScript:           Указывает, следует ли сохранить скомпилированный скрипт.
        :param executionContextId:      (optional) Указывает, в каком контексте выполнения выполнять сценарий.
                                            Если параметр не указан, выражение будет выполняться в контексте
                                            проверяемой страницы.
        :return:                        {
                                            "scriptId": str()
                                            "exceptionDetails": dict(https://chromedevtools.github.io/devtools-protocol/tot/Runtime#type-ExceptionDetails)
                                        }
        """
        args = {"expression": expression, "sourceURL": sourceURL, "persistScript": persistScript}
        if executionContextId is not None:
            args.update({"executionContextId": executionContextId})

        response = await self.Call("Runtime.compileScript", args)
        if "exceptionDetails" in response:
            raise Exception(response["exceptionDetails"]["text"] + "\n" + json.dumps(response["exceptionDetails"]))
        return response["scriptId"]

    async def BuildScript(self, expression: str, context: Optional['RuntimeType.ContextDescription'] = None) -> 'Script':
        return Script(self, expression, context)

    async def RunIfWaitingForDebugger(self) -> None:
        """
        Сообщает инспектируемой странице, что можно запуститься, если она ожидает этого после
            Target.setAutoAttach.
        """
        await self.Call("Runtime.runIfWaitingForDebugger")

    async def RunScript(
            self, scriptId: str,
            executionContextId: Optional[int] = None,
            objectGroup: str = "console",
            silent: bool = False,
            includeCommandLineAPI: bool = True,
            returnByValue: bool = False,
            generatePreview: bool = False,
            awaitPromise: bool = True
    ) -> dict:
        """
        Запускает скрипт с заданным идентификатором в заданном контексте.
        https://chromedevtools.github.io/devtools-protocol/tot/Runtime#method-runScript
        :param scriptId:                ID сценария для запуска.
        :param executionContextId:      (optional) Указывает, в каком контексте выполнения выполнять сценарий.
                                            Если параметр не указан, выражение будет выполняться в контексте
                                            проверяемой страницы.
        :param objectGroup:             (optional) Символическое имя группы, которое можно использовать для
                                            освобождения нескольких объектов.
        :param silent:                  (optional) В тихом режиме исключения, выданные во время оценки, не
                                            сообщаются и не приостанавливают выполнение. Переопределяет
                                            состояние setPauseOnException.
        :param includeCommandLineAPI:   (optional) Определяет, должен ли API командной строки быть доступным
                                            во время оценки.
        :param returnByValue:           (optional) Ожидается ли результат в виде объекта JSON, который должен
                                            быть отправлен по значению.
        :param generatePreview:         (optional) Должен ли предварительный просмотр генерироваться для результата.
        :param awaitPromise:            (optional) Будет ли выполнено ожидание выполнения для полученного значения
                                            и возврата после ожидаемого 'promise'.
        :return:                        {
                                            "result": dict(https://chromedevtools.github.io/devtools-protocol/tot/Runtime#type-RemoteObject)
                                            "exceptionDetails": dict(https://chromedevtools.github.io/devtools-protocol/tot/Runtime#type-ExceptionDetails)
                                        }
        """
        args = {
            "scriptId": scriptId, "objectGroup": objectGroup, "silent": silent,
            "includeCommandLineAPI": includeCommandLineAPI, "returnByValue": returnByValue,
            "generatePreview": generatePreview, "awaitPromise": awaitPromise
        }
        if executionContextId is not None:
            args.update({"executionContextId": executionContextId})

        response = await self.Call("Runtime.runScript", args)
        if "exceptionDetails" in response:
            raise Exception(response["result"]["description"] + "\n" + json.dumps(response["exceptionDetails"]))
        return response["result"]

    async def AddBinding(self, name: str, executionContextName: Optional[int] = None) -> None:
        """
        (EXPERIMENTAL)
        Если executeContextId пуст, добавляет привязку с заданным именем к глобальным объектам всех
            проверенных контекстов, включая созданные позже, привязки переживают перезагрузки. Если
            указан executeContextId, добавляет привязку только к глобальному объекту данного
            контекста выполнения. Функция привязки принимает ровно один аргумент, этот аргумент
            должен быть строкой, в случае любого другого ввода функция выдает исключение. Каждый
            вызов функции привязки создает уведомление Runtime.bindingCalled.
        https://chromedevtools.github.io/devtools-protocol/tot/Runtime#method-addBinding
        :param name:                    Имя привязки.
        :param executionContextName:      (optional) Идентификатор контекста исполнения.
        :return:
        """
        args = {"name": name}
        if executionContextName is not None:
            args.update({"executionContextName": executionContextName})

        await self.Call("Runtime.addBinding", args)

    @abstractmethod
    async def Call(
            self, domain_and_method: str,
            params: Optional[dict] = None,
            wait_for_response: bool = True
    ) -> Union[dict, None]: raise NotImplementedError("async method Call() — is not implemented")

    @abstractmethod
    async def AddListenerForEvent(
            self, event: Union[str, DomainEvent], listener: Callable, *args: any) -> None:
        raise NotImplementedError("async method AddListenerForEvent() — is not implemented")

    @abstractmethod
    def RemoveListenerForEvent(self, event: Union[str, DomainEvent], listener: Callable) -> None:
        raise NotImplementedError("async method RemoveListenerForEvent() — is not implemented")


class RuntimeEvent(DomainEvent):
    consoleAPICalled = "Runtime.consoleAPICalled"
    exceptionRevoked = "Runtime.exceptionRevoked"
    exceptionThrown = "Runtime.exceptionThrown"
    executionContextCreated = "Runtime.executionContextCreated"
    executionContextDestroyed = "Runtime.executionContextDestroyed"
    executionContextsCleared = "Runtime.executionContextsCleared"
    inspectRequested = "Runtime.inspectRequested"
    bindingCalled = "Runtime.bindingCalled"                       # ! EXPERIMENTAL


class RuntimeType:

    @dataclass
    class PropertyPreview:
        name: str
        type: Literal["object", "function", "undefined", "string", "number", "boolean", "symbol", "accessor", "bigint"]
        valuePreview: Optional['RuntimeType.ObjectPreview']
        value: Optional[str] = None
        _valuePreview: Optional['RuntimeType.ObjectPreview'] = field(init=False, repr=False, default=None)
        subtype: Literal[
            "array", "null", "node", "regexp", "date", "map", "set", "weakmap", "weakset", "iterator", "generator",
            "error", "proxy", "promise", "typedarray", "arraybuffer", "dataview", "webassemblymemory", "wasmvalue"] = None

        @property
        def valuePreview(self) -> 'RuntimeType.ObjectPreview':
            return self._valuePreview

        @valuePreview.setter
        def valuePreview(self, data: dict) -> None:
            self._valuePreview = RuntimeType.ObjectPreview(**data) if not isinstance(data, property) else None


    @dataclass
    class EntryPreview:
        _value: 'RuntimeType.ObjectPreview'
        _key: Optional['RuntimeType.ObjectPreview']

        @property
        def key(self) -> 'RuntimeType.ObjectPreview':
            return self._key

        @key.setter
        def key(self, data: dict) -> None:
            self._key = RuntimeType.ObjectPreview(**data)

        @property
        def value(self) -> 'RuntimeType.ObjectPreview':
            return self._value

        @value.setter
        def value(self, data: dict) -> None:
            self._value = RuntimeType.ObjectPreview(**data)


    @dataclass
    class CustomPreview:
        header: str
        bodyGetterId: Optional[str] = None


    @dataclass
    class ObjectPreview:
        type: Literal["object", "function", "undefined", "string", "number", "boolean", "symbol", "bigint"]
        overflow: bool
        properties: List['RuntimeType.PropertyPreview']
        entries: List['RuntimeType.EntryPreview']
        subtype: Optional[Literal[
            "array", "null", "node", "regexp", "date", "map", "set", "weakmap", "weakset", "iterator", "generator",
            "error", "proxy", "promise", "typedarray", "arraybuffer", "dataview", "webassemblymemory", "wasmvalue"]] = None
        description: Optional[str] = None
        _entries: List['RuntimeType.EntryPreview'] = field(init=False, repr=False, default=None)

        @property
        def properties(self) -> List['RuntimeType.PropertyPreview']:
            return self._properties

        @properties.setter
        def properties(self, data: List[dict]) -> None:
            self._properties = [RuntimeType.PropertyPreview(**item) for item in data]

        @property
        def entries(self) -> List['RuntimeType.EntryPreview']:
            return self._entries

        @entries.setter
        def entries(self, data: List[dict]) -> None:
            self._entries = [RuntimeType.EntryPreview(**item) for item in data] if not isinstance(data, property) else None


    @dataclass
    class RemoteObject:
        """
        Зеркальный объект, ссылающийся на исходный объект JavaScript.
        # https://chromedevtools.github.io/devtools-protocol/tot/Runtime/#type-RemoteObject
        """
        type: Literal["object", "function", "undefined", "string", "number", "boolean", "symbol", "bigint"]
        preview: Optional['RuntimeType.ObjectPreview']
        customPreview: Optional['RuntimeType.CustomPreview']
        subtype: Optional[Literal[
            "array", "null", "node", "regexp", "date", "map", "set", "weakmap", "weakset", "iterator", "generator",
            "error", "proxy", "promise", "typedarray", "arraybuffer", "dataview", "webassemblymemory", "wasmvalue"]] = None
        className: Optional[str] = None
        value: Optional[any] = None
        unserializableValue: Optional[str] = None   # ? Примитивное значение, которое не может быть преобразовано в строку
                                                    # ?     JSON, не имеет value, но получает это свойство.
        description: Optional[str] = None
        objectId: Optional[str] = None
        _preview: Optional['RuntimeType.ObjectPreview'] = field(init=False, repr=False, default=None)
        _customPreview: Optional['RuntimeType.CustomPreview'] = field(init=False, repr=False, default=None)

        @property
        def preview(self) -> 'RuntimeType.ObjectPreview':
            return self._preview

        @preview.setter
        def preview(self, data: dict) -> None:
            self._vpreview = RuntimeType.ObjectPreview(**data) if not isinstance(data, property) else None

        @property
        def customPreview(self) -> 'RuntimeType.CustomPreview':
            return self._customPreview

        @customPreview.setter
        def customPreview(self, data: dict) -> None:
            self._customPreview = RuntimeType.CustomPreview(**data) if not isinstance(data, property) else None


    @dataclass
    class AuxData:
        isDefault: bool
        type: str
        frameId: str


    @dataclass
    class ContextDescription:
        id: int
        origin: str     # url
        name: str
        uniqueId: str
        auxData: 'RuntimeType.AuxData'

        @property
        def auxData(self) -> 'RuntimeType.AuxData':
            return self._auxData

        @auxData.setter
        def auxData(self, data: dict) -> None:
            self._auxData = RuntimeType.AuxData(**data)

    @dataclass
    class PropertyDescriptor:
        name: str
        configurable: bool
        enumerable: bool
        value: Optional['RuntimeType.RemoteObject'] = None
        writable: Optional[bool] = None
        get: Optional['RuntimeType.RemoteObject'] = None
        set: Optional['RuntimeType.RemoteObject'] = None
        wasThrown: Optional[bool] = None
        isOwn: Optional[bool] = None
        symbol: Optional['RuntimeType.RemoteObject'] = None



class Script:
    """
    Упаковывает вызываемое в контексте указанного фрейма выражение. Если контекст не указан, в его качестве
        будет выбран фрейм верхнего уровня. Выражение можно заменить при вызове.
    """
    def __init__(
            self, page_instance, expression: str,
            context: Optional[Union['RuntimeType.ContextDescription', str]] = None):
        self.page_instance = page_instance
        self.expression = expression
        self.unique_context_id = context if type(context) is str else context.uniqueId if context is not None else None

    async def Call(
            self, expression: Optional[str] = None, returnByValue: Optional[bool] = None) -> 'RuntimeType.RemoteObject':
        if expression:
            self.expression = expression
        args = {"expression": self.expression}
        if returnByValue is not None:
            args.update(returnByValue=returnByValue)
        if self.unique_context_id is not None:
            args.update(uniqueContextId=self.unique_context_id)
        result: dict = await self.page_instance.Call("Runtime.evaluate", args)
        return RuntimeType.RemoteObject(**result.get("result"))


class ContextManager:
    contexts: List['RuntimeType.ContextDescription'] = []
    is_watch: bool = False

    async def on_create(self, data: dict) -> None:
        self.contexts.append(RuntimeType.ContextDescription(**data.get("context")))

    async def on_clear(self, data: dict) -> None:
        self.contexts = []

    async def on_destroy(self, data: dict) -> None:
        context_id: int = data.get("executionContextId")
        ii = -1
        for i, ctx in enumerate(self.contexts):
            if ctx.id == context_id:
                ii = i
                break

        if ii > -1: self.contexts.pop(ii)

    def GetDefaultContext(self, frameId: str) -> Union['RuntimeType.ContextDescription', None]:
        for ctx in self.contexts:
            print("frameId", frameId)
            print("Runtime context", ctx, len(self.contexts))
            if ctx.auxData.frameId == frameId and ctx.auxData.isDefault:
                return ctx
        return None


from abc import ABC, abstractmethod
from typing import Optional, Union

class Console(ABC):
    """
    #   Этот домен является устаревшим. Используйте Runtime или Log вместо него.
    #   https://chromedevtools.github.io/devtools-protocol/tot/Console/
    """
    __slots__ = ()

    def __init__(self):
        self.console_domain_enabled   = False

    @property
    def connected(self) -> bool:
        return False

    @property
    def verbose(self) -> bool:
        return False

    @property
    def page_id(self) -> str:
        return ""

    async def ConsoleEnable(self) -> None:
        """
        Включает у домена 'Console', отправку клиенту собранные на данный момент сообщения посредством
            'messageAdded' уведомления.
        https://chromedevtools.github.io/devtools-protocol/tot/Console#method-enable
        :return:
        """
        await self.Call("Console.enable")
        self.console_domain_enabled = True

    async def ConsoleDisable(self) -> None:
        """
        Выключает 'Console' домен, останавливая отправку сообщений.
        https://chromedevtools.github.io/devtools-protocol/tot/Console#method-disable
        :return:
        """
        await self.Call("Console.disable")
        self.console_domain_enabled = False

    async def ClearConsole(self) -> None:
        """
        Ничего не делает.
        https://chromedevtools.github.io/devtools-protocol/tot/Console#method-clearMessages
        :return:
        """
        await self.Call("Console.clearMessages")

    @abstractmethod
    async def Call(
            self, domain_and_method: str,
            params: Optional[dict] = None,
            wait_for_response: bool = True
    ) -> Union[dict, None]: raise NotImplementedError("async method Call() — is not implemented")



from abc import ABC, abstractmethod
from typing import Optional, Union

class DeviceOrientation(ABC):
    """
    #   https://chromedevtools.github.io/devtools-protocol/tot/DeviceOrientation
    """
    __slots__ = ()

    @property
    def connected(self) -> bool:
        return False

    @property
    def verbose(self) -> bool:
        return False

    @property
    def page_id(self) -> str:
        return ""

    async def ClearDeviceOrientationOverride(self) -> None:
        """
        Очищает переопределенную ориентацию устройства.
        https://chromedevtools.github.io/devtools-protocol/tot/DeviceOrientation/#method-clearDeviceOrientationOverride
        :return:
        """
        await self.Call("DeviceOrientation.clearDeviceOrientationOverride")

    async def SetDeviceOrientationOverride(self, alpha: float, beta: float, gamma: float) -> None:
        """
        Переопределяет ориентацию устройства, принудительно изменяя значения сенсоров, котоые так же
            можно найти в консоли браузера по Ctrl+Shift+P и в поиске ввести 'Show Sensors'.
        https://chromedevtools.github.io/devtools-protocol/tot/DeviceOrientation/#method-setDeviceOrientationOverride
        :return:
        """
        args = {"alpha": alpha, "beta": beta, "gamma": gamma}
        await self.Call("DeviceOrientation.setDeviceOrientationOverride", args)

    @abstractmethod
    async def Call(
            self, domain_and_method: str,
            params: Optional[dict] = None,
            wait_for_response: bool = True
    ) -> Union[dict, None]: raise NotImplementedError("async method Call() — is not implemented")



import asyncio
from asyncio import sleep, wait_for
from urllib.error import URLError
from inspect import iscoroutinefunction

from .Data import CommonCallback
from .Browser import Browser
from .PageEx import PageEx
from .exceptions import NoTargetWithGivenIdFound

from urllib.parse import urlparse
import re
from typing import Optional, List


class BrowserEx(Browser):
    """
    Расширение для 'Browser'. Включает в себя один новый метод CreatePage()
        создающий новые вкладки(процессы страниц), а так же, все методы
        получения инстансов вкладок, возвращают их с расширенным функционалом.
    """

    @staticmethod
    def get_domain(url):
        if url == "about:blank":
            return "about:blank"
        return re.search(r"^[^.]+", urlparse(url).netloc)[0]

    async def GetPageBy(
            self, key: str,
            value: str,
            match_mode: str = "exact",
            index: int = 0,
            callback: CommonCallback = None
    ) -> Optional[PageEx]:

        if callback is not None and not iscoroutinefunction(callback):
            raise TypeError("Argument 'callback' must be a coroutine")

        counter = 0; v = value.lower()
        for page_data in await self.GetPageList():
            data = page_data[key].lower()
            if ((match_mode == "exact" and data == v)
                or (match_mode == "contains" and data.find(v) > -1)
                    or (match_mode == "startswith" and data.find(v) == 0)):
                if counter == index:
                    page = PageEx(
                        page_data["webSocketDebuggerUrl"],
                        page_data["id"],
                        page_data["devtoolsFrontendUrl"],
                        callback,
                        self.profile_path == "",
                        self.verbose,
                        self.browser_name
                    )
                    await page.Activate()
                    return page
                counter += 1
        return None

    async def GetPage(
            self, index: int = 0, page_type: str = "page",
            callback: CommonCallback = None
    ) -> Optional[PageEx]:
        return await self.GetPageBy("type", page_type, "exact", index, callback)

    async def GetPageByID(
            self, page_id: str, index: int = 0,
            callback: CommonCallback = None
    ) -> Optional[PageEx]:
        return await self.GetPageBy("id", page_id, "exact", index, callback)

    async def GetPageByTitle(
            self, value: any, match_mode: str = "startswith",
            index: int = 0, callback: CommonCallback = None
    ) -> Optional[PageEx]:
        return await self.GetPageBy("title", value, match_mode, index, callback)

    async def GetPageByURL(
            self, value: any, match_mode: str = "startswith",
            index: int = 0, callback: CommonCallback = None
    ) -> Optional[PageEx]:
        return await self.GetPageBy("url", value, match_mode, index, callback)

    async def CreatePage(
            self,  url: str = "about:blank",
            newWindow: bool = False,
            background: bool = False,
            wait_for_create: bool = True,
            callback: CommonCallback = None
    ) -> Optional[PageEx]:
        """
        Создаёт новую вкладку в браузере.
        :param url:                     - (optional) Адрес будет открыт при создании.
        :param newWindow:               - (optional) Если 'True' — страница будет открыта в новом окне.
        :param background:              - (optional) Если 'True' — страница будет открыта в фоне.
        :return:                    * Инстанс страницы <PageEx>
        """
        while not (tmp := await self.GetPage(callback=callback)):
            await sleep(.5)
        page_id = await tmp.CreateTarget(url, newWindow=newWindow, background=background)
        if wait_for_create:
            while not(page := await self.GetPageByID(page_id)):
                await sleep(.5)
        else:
            page = await self.GetPageByID(page_id)
        return page

    async def ShowInspector(self, page: PageEx, new_window: bool = True,
                            callback: CommonCallback = None) -> Optional[PageEx]:
        """
        Открывает новую вкладку с дебаггером для инспектируемой страницы.
        :param page:            - Инспектируемая страница. Может принадлежать любому браузеру.
        :param new_window:      - Создать target в отдельном окне?
        :return:        <PageEx>
        """
        return await self.CreatePage(
            "http://127.0.0.1:" + str(self.debug_port) + page.frontend_url, new_window, callback=callback
        )

    async def CreatePopupWindow(self, page: PageEx, url: str = "about:blank",
                                callback: CommonCallback = None) -> Optional[PageEx]:
        """
        Создаёт всплывающее окно с минимумом интерфейса браузера".
        :param url:             - Адрес, ресурс которого будет загружен
        :param page:            - Родительская страница, инициатор
        :return:        PageEx or None
        """
        await page.InjectJS(f'window.open("{url}", "blank_window_name", "popup,noopener,noreferrer");')
        return await self.GetPageByOpener(page, callback=callback)

    async def GetPageByOpener(self, page: PageEx, callback: CommonCallback = None) -> Optional[PageEx]:
        """
        Возвращает последний созданный инстанс страницы, открытие которого инициировано с конкретной страницы.
            Например, при использовании JavaScript "window.open()".
        :param page:            - Родительская страница, инициатор
        :return:        PageEx or None
        """
        for target_info in await page.GetTargets():
            if target_info.openerId == page.page_id:
                return await self.GetPageByID(target_info.targetId, callback=callback)
        return None

    async def GetPagesByOpener(self, page: PageEx, callback: CommonCallback = None) -> List[PageEx]:
        """
        Возвращает список всех инстансов страниц, открытие которых инициировано с конкретной страницы.
            Например, при использовании JavaScript "window.open()".
        :param page:            - Родительская страница, инициатор открытых окон
        :return:        List[PageEx]
        """
        pages = []
        for target_info in await page.GetTargets():
            if target_info.openerId == page.page_id:
                pages.append(await self.GetPageByID(target_info.targetId, callback=callback))
        return pages

    async def WaitFirstTab(self, timeout: float = 20.0, callback: CommonCallback = None) -> PageEx:
        """
        Вызывает исключение 'asyncio.exceptions.TimeoutError' по истечении таймаута, или возвращает инстанс.
        """
        return await wait_for(self.GetFirstTab(callback), timeout)

    async def GetFirstTab(self, callback: CommonCallback = None) -> PageEx:
        """
        Безусловно дожидается соединения со страницей.
        """
        while True:
            try:
                while (page := await self.GetPage(callback=callback)) is None:
                    await sleep(.5)
                return page
            except URLError: await sleep(1)

    async def Close(self) -> None:
        """ Корректно закрывает браузер если остались ещё его инстансы """
        if page := await self.GetPage():
            await page.CloseBrowser()

    async def CloseAllPagesExcept(self, *except_list: PageEx) -> None:
        """ Закрывает все страницы браузера, кроме переданных """
        for conn in await self.GetTargetConnectionInfoList():
            if conn.type == "page":
                condition = False
                for page in except_list:
                    condition |= page.page_id == conn.id
                if not condition:
                    i = 4
                    try:
                        while (tab := await self.GetPageByID(conn.id)) is None and i:
                            await asyncio.sleep(.5)
                            i -= 1
                        if tab: await tab.Close()
                    except NoTargetWithGivenIdFound:
                        pass

    async def GetFramesFor(self, page: PageEx) -> List[PageEx]:
        """ Возвращает список iFrame для указанного инстанса """
        result = []
        for conn in await self.GetPageList():
            if conn["type"] == "iframe" and conn["parentId"] == page.page_id:
                result.append( await self.GetPageByID(conn["id"]) )
        return result

    def __eq__(self, other: "BrowserEx") -> bool:
        return self.debug_port == other.debug_port

    def __hash__(self) -> int:
        return hash(self.debug_port)



from abc import ABC, abstractmethod
from typing import Optional, Union, Callable, List
from ..Data import DomainEvent

class CSS(ABC):
    """
    #   https://chromedevtools.github.io/devtools-protocol/tot/Log
    #   LogEntry -> https://chromedevtools.github.io/devtools-protocol/tot/Log#type-LogEntry
    """
    __slots__ = ()

    def __init__(self):
        self.css_domain_enabled = False
        self.style_sheets = []  # Если домен CSS активирован, сюда попадут все 'styleSheetId' страницы

    @property
    def connected(self) -> bool:
        return False

    @property
    def verbose(self) -> bool:
        return False

    @property
    def page_id(self) -> str:
        return ""

    @property
    def style_sheets(self) -> list:
        return self._style_sheets

    @style_sheets.setter
    def style_sheets(self, value) -> None:
        self._style_sheets = value

    async def _CSS_sheet_catcher(self, data: dict) -> None:
        """
        Колбэк вызываемый для каждого 'CSS.styleSheetAdded'-события, если
            включён агент домена 'CSS'.
        """
        self.style_sheets.append(data["header"]["styleSheetId"])

    async def CSSEnable(self) -> None:
        """
        Включает агент CSS. Клиент не должен предполагать, что агент CSS включен,
            пока не будет получен результат этой команды.
        https://chromedevtools.github.io/devtools-protocol/tot/CSS/#method-disable
        :return:
        """
        if not self.css_domain_enabled:
            await self.AddListenerForEvent("CSS.styleSheetAdded", self._CSS_sheet_catcher)
            await self.Call("CSS.enable")
            self.css_domain_enabled = True

    async def CSSDisable(self) -> None:
        """
        Отключает агент CSS.
        https://chromedevtools.github.io/devtools-protocol/tot/CSS/#method-disable
        :return:
        """
        if self.css_domain_enabled:
            self.RemoveListenerForEvent("CSS.styleSheetAdded", self._CSS_sheet_catcher)
            await self.Call("CSS.disable")
            self.css_domain_enabled = False

    async def AddRule(
        self, styleSheetId: str, ruleText: str,
        startLine:   int = 0,
        startColumn: int = 0,
        endLine:     int = 0,
        endColumn:   int = 0
    ) -> dict:
        """
        Вставляет новое правило с заданным ruleText в таблицу стилей с заданным styleSheetId в позицию,
            указанную с помощью location.
        https://chromedevtools.github.io/devtools-protocol/tot/CSS/#method-addRule
        :param styleSheetId:            Идентификатор стилей.
        :param ruleText:                Текст правила (например: 'body { background: red }').

        Текстовый диапазон в ресурсе. Все числа отсчитываются от нуля.
        :param startLine:               (optional) - Стартовая линия диапазона.
        :param startColumn:             (optional) - Начальный столбец диапазона (включительно).
        :param endLine:                 (optional) - Конечная строка диапазона.
        :param endColumn:               (optional) - Конечный столбец диапазона (исключая).
        :return:    https://chromedevtools.github.io/devtools-protocol/tot/CSS/#type-CSSRule
        """
        location = {"startLine": startLine, "startColumn": startColumn, "endLine": endLine, "endColumn": endColumn}
        args = {
            "styleSheetId": styleSheetId,
            "ruleText": ruleText,
            "location": location
        }
        return (await self.Call("CSS.addRule", args))["rule"]

    async def CollectClassNames(self, styleSheetId: str) -> List[str]:
        """
        Возвращает список всех имён классов для указанного ID.
        https://chromedevtools.github.io/devtools-protocol/tot/CSS/#method-collectClassNames
        :param styleSheetId:            Идентификатор стилей.
        :return:
        """
        return (await self.Call("CSS.collectClassNames", {"styleSheetId": styleSheetId}))["classNames"]

    async def CreateStyleSheet(self, frameId: str = None) -> str:
        """
        Создает новую специальную таблицу стилей через "инспектор" во фрейме с заданным frameId.
            Если frameId не указан, будет использован frameId главного фрейма текущей страницы.
            Возвращает ID созданного styleSheet.
        https://chromedevtools.github.io/devtools-protocol/tot/CSS/#method-createStyleSheet
        :param frameId:                 Идентификатор фрейма, в котором будет создан styleSheet.
        :return:                styleSheetId
        """
        frameId = frameId if frameId else self.page_id
        styleSheetId = (await self.Call("CSS.createStyleSheet", {"frameId": frameId}))["styleSheetId"]
        self.style_sheets.append(styleSheetId)
        return styleSheetId

    async def GetStyleSheetText(self, styleSheetId: str) -> str:
        """
        Возвращает текстовый контент стилей для указанного ID.
        https://chromedevtools.github.io/devtools-protocol/tot/CSS/#method-getStyleSheetText
        :param styleSheetId:            Идентификатор стилей.
        :return:
        """
        return (await self.Call("CSS.getStyleSheetText", {"styleSheetId": styleSheetId}))["text"]

    @abstractmethod
    async def AddListenerForEvent(
            self, event: str, listener: Callable, *args: any) -> None:
        raise NotImplementedError("async method AddListenerForEvent() — is not implemented")

    @abstractmethod
    def RemoveListenerForEvent(self, event: str, listener: Callable) -> None:
        raise NotImplementedError("method RemoveListenerForEvent() — is not implemented")

    @abstractmethod
    async def Call(
            self, domain_and_method: str,
            params: Optional[dict] = None,
            wait_for_response: bool = True
    ) -> Union[dict, None]: raise NotImplementedError("async method Call() — is not implemented")

class CSSEvent(DomainEvent):
    fontsUpdated = "CSS.fontsUpdated"
    mediaQueryResultChanged = "CSS.mediaQueryResultChanged"
    styleSheetAdded = "CSS.styleSheetAdded"
    styleSheetChanged = "CSS.styleSheetChanged"
    styleSheetRemoved = "CSS.styleSheetRemoved"



import asyncio
import sys, re
import urllib.request
if sys.platform == "win32":
    import winreg


def get_request(url: str) -> str:
    with urllib.request.urlopen(url) as response:
        return response.read().decode('utf-8')


def registry_read_key(exe="chrome") -> str:
    """ Возвращает путь до EXE.
    """
    reg_path = f"HKEY_LOCAL_MACHINE\\SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\App Paths\\{exe}.exe"
    key, path = re.findall(r"(^[^\\/]+)[\\/](.*)", reg_path)[0]
    connect_to = eval(f"winreg.{key}")
    try: h_key = winreg.OpenKey( winreg.ConnectRegistry(None, connect_to), path )
    except FileNotFoundError: return ""
    result = winreg.QueryValue(h_key, None)
    winreg.CloseKey(h_key)
    return result


def log(data: any = "", lvl: str = "[<- V ->]", eol: str = "\n") -> None:
    print(f"\x1b[32m{lvl} \x1b[38m\x1b[3m{data}\x1b[0m", end=eol)


def save_img_as(path: str, data: bytes)-> None:
    """ Сохраняет по пути 'path' набор байт 'data', которые можно прислать
    из метода страницы MakeScreenshot()
    """
    with open(path, "wb") as f:
        f.write(data)


async def async_util_call(function: callable, *args) -> any:
    """ Позволяет выполнять неблокирующий вызов блокирующих функций. Например:
    await async_util_call(
        save_img_as, "ScreenShot.png", await page_instance.MakeScreenshot()
    )
    """
    return await asyncio.get_running_loop().run_in_executor(
        None, function, *args
    )



from abc import ABC, abstractmethod
from typing import Optional, Union
from ..Data import DomainEvent

class BackgroundService(ABC):
    """
    #   https://chromedevtools.github.io/devtools-protocol/tot/BackgroundService
    """
    __slots__ = ()

    def __init__(self):
        self.observing_started = False
        self.recording_started = False

    @property
    def connected(self) -> bool:
        return False

    @property
    def verbose(self) -> bool:
        return False

    @property
    def page_id(self) -> str:
        return ""

    async def BackgroundClearEvents(self, service: str) -> None:
        """
        Удаляет все сохраненные данные для service.
        https://chromedevtools.github.io/devtools-protocol/tot/BackgroundService/#method-clearEvents
        :param service:         Фоновая служба, которая будет связана с командами / событиями. Каждая фоновая
                                    служба работает независимо, но использует один и тот же API.
                                Допустимые значения:
                                    backgroundFetch, backgroundSync, pushMessaging, notifications,
                                    paymentHandler,periodicBackgroundSync
        :return:
        """
        await self.Call("BackgroundService.clearEvents", {"service": service})

    async def BackgroundSetRecording(self, shouldRecord: bool, service: str) -> None:
        """
        Включает, или выключает запись для service.
        https://chromedevtools.github.io/devtools-protocol/tot/BackgroundService/#method-setRecording
        :param shouldRecord:    True == включить
        :param service:         ---------------------
        :return:
        """
        await self.Call("BackgroundService.clearEvents", {"shouldRecord": shouldRecord, "service": service})
        self.recording_started = shouldRecord

    async def BackgroundStartObserving(self, service: str) -> None:
        """
        Включает обновления событий для service.
        https://chromedevtools.github.io/devtools-protocol/tot/BackgroundService/#method-startObserving
        :param service:         ---------------------
        :return:
        """
        if not self.observing_started:
            await self.Call("BackgroundService.startObserving", {"service": service})
            self.observing_started = True

    async def BackgroundStopObserving(self, service: str) -> None:
        """
        Выключает обновления событий для service.
        https://chromedevtools.github.io/devtools-protocol/tot/BackgroundService/#method-stopObserving
        :param service:         ---------------------
        :return:
        """
        if self.observing_started:
            await self.Call("BackgroundService.stopObserving", {"service": service})
            self.observing_started = False

    @abstractmethod
    async def Call(
            self, domain_and_method: str,
            params: Optional[dict] = None,
            wait_for_response: bool = True
    ) -> Union[dict, None]: raise NotImplementedError("async method Call() — is not implemented")

class BackgroundServiceEvent(DomainEvent):
    backgroundServiceEventReceived = "BackgroundService.backgroundServiceEventReceived"
    recordingStateChanged = "BackgroundService.recordingStateChanged"



from abc import ABC, abstractmethod
from typing import Optional, Union
from ..Data import DomainEvent

class Overlay(ABC):
    """
    #   https://chromedevtools.github.io/devtools-protocol/tot/Overlay
    """
    __slots__ = ()

    def __init__(self):
        self.overlay_domain_enabled = False

    @property
    def connected(self) -> bool:
        return False

    @property
    def verbose(self) -> bool:
        return False

    @property
    def page_id(self) -> str:
        return ""

    async def OverlayEnable(self) -> None:
        """
        Включает отправку уведомлений домена 'Overlay'.
        https://chromedevtools.github.io/devtools-protocol/tot/Overlay#method-enable
        :return:
        """
        await self.Call("Overlay.enable")
        self.overlay_domain_enabled = True

    async def OverlayDisable(self) -> None:
        """
        Выключает 'Overlay' домен, останавливая отправку сообщений.
        https://chromedevtools.github.io/devtools-protocol/tot/Overlay#method-disable
        :return:
        """
        await self.Call("Overlay.disable")
        self.overlay_domain_enabled = False

    async def SetShowFPSCounter(self) -> None:
        """
        Запрашивает бэкэнд показ счетчика FPS.
        https://chromedevtools.github.io/devtools-protocol/tot/Overlay#method-setShowFPSCounter
        :return:
        """
        await self.Call("Overlay.setShowFPSCounter")

    @abstractmethod
    async def Call(
            self, domain_and_method: str,
            params: Optional[dict] = None,
            wait_for_response: bool = True
    ) -> Union[dict, None]: raise NotImplementedError("async method Call() — is not implemented")

class OverlayEvent(DomainEvent):
    inspectModeCanceled = "Overlay.inspectModeCanceled"
    inspectNodeRequested = "Overlay.inspectNodeRequested"
    nodeHighlightRequested = "Overlay.nodeHighlightRequested"
    screenshotRequested = "Overlay.screenshotRequested"



from abc import ABC, abstractmethod
from typing import Optional, Union
from ..Data import DomainEvent

class Emulation(ABC):
    """
    #   https://chromedevtools.github.io/devtools-protocol/tot/Emulation
    """
    __slots__ = ()

    @property
    def connected(self) -> bool:
        return False

    @property
    def verbose(self) -> bool:
        return False

    @property
    def page_id(self) -> str:
        return ""

    async def CanEmulate(self) -> bool:
        """
        Сообщает, поддерживается ли эмуляция.
        https://chromedevtools.github.io/devtools-protocol/tot/Emulation#method-canEmulate
        :return:                result -> True если эмуляция поддерживается
        """
        return (await self.Call("Emulation.canEmulate"))["result"]

    async def ClearDeviceMetricsOverride(self) -> None:
        """
        Очищает переопределённые метрики устройства.
        https://chromedevtools.github.io/devtools-protocol/tot/Emulation#method-clearDeviceMetricsOverride
        :return:
        """
        await self.Call("Emulation.clearDeviceMetricsOverride")

    async def ClearGeolocationOverride(self) -> None:
        """
        Очищает переопределённые позицию геолокации и ошибку.
        https://chromedevtools.github.io/devtools-protocol/tot/Emulation#method-clearGeolocationOverride
        :return:
        """
        await self.Call("Emulation.clearGeolocationOverride")

    async def ResetPageScaleFactor(self) -> None:
        """
        Запрашивает сброс масштабирования страницы до начальных значений.
        https://chromedevtools.github.io/devtools-protocol/tot/Emulation#method-resetPageScaleFactor
        :return:
        """
        await self.Call("Emulation.resetPageScaleFactor")

    async def SetFocusEmulationEnabled(self, enabled: bool) -> None:
        """
        Включает или отключает симуляцию фокуса(когда страница активна).
        https://chromedevtools.github.io/devtools-protocol/tot/Emulation#method-setFocusEmulationEnabled
        :param enabled:         Включает, или отключает эмуляцию фокуса.
        :return:
        """
        await self.Call("Emulation.setFocusEmulationEnabled", {"enabled": enabled})

    async def SetCPUThrottlingRate(self, rate: float) -> None:
        """
        Включает CPU "throttling", эмулируя медленный процессор.
        https://chromedevtools.github.io/devtools-protocol/tot/Emulation#method-setCPUThrottlingRate
        :param rate:            Коэффициент замедления(1 - без замедления; 2 - 2х кратное, и т.д.).
        :return:
        """
        await self.Call("Emulation.setCPUThrottlingRate", {"rate": rate})

    async def SetDefaultBackgroundColorOverride(self, color: dict) -> None:
        """
        Устанавливает или очищает переопределение цвета фона фрейма по умолчанию. Это переопределение
            используется, если в содержимом оно не указано.
        https://chromedevtools.github.io/devtools-protocol/tot/Emulation#method-setDefaultBackgroundColorOverride
        :param color:            (optional) RGBA цвета фона по умолчанию. Если не указано, любое
                                    существующее переопределение будет очищено. Словарь вида:
                                        {
                                            "r": int[0-255],
                                            "g": int[0-255],
                                            "b": int[0-255],
                                            "a": float[0-1] -> опционально. По умолчанию: 1.
                                        }
        :return:
        """
        await self.Call("Emulation.setDefaultBackgroundColorOverride", {"color": color})

    async def SetDeviceMetricsOverride(
            self, width: int, height: int,
            deviceScaleFactor: float = 0,
                        mobile: bool = False,
                        scale: Optional[float] = None,
                    screenWidth: Optional[int] = None,
                   screenHeight: Optional[int] = None,
                      positionX: Optional[int] = None,
                      positionY: Optional[int] = None,
            dontSetVisibleSize: Optional[bool] = None,
             screenOrientation: Optional[dict] = None,
                      viewport: Optional[dict] = None
    ) -> None:
        """
        Переопределяет значения размеров экрана устройства (результаты мультимедийного запроса CSS,
            относящиеся к «device-width» / «device-height», связанные с window.screen.width,
            window.screen.height, window.innerWidth, window.innerHeight).
        https://chromedevtools.github.io/devtools-protocol/tot/Emulation#method-setDeviceMetricsOverride
        :param width:               Переопределяет значение ширины viewport в пикселях(от 0 до 10_000_000).
                                        window.innerWidth
                                        0 — выключает переопределение.
        :param height:              Переопределяет значение высоты viewport в пикселях(от 0 до 10_000_000).
                                        window.innerHeight
                                        0 — выключает переопределение.
        :param deviceScaleFactor:   Переопределяет значение масштабирования устройства. Отношение размера
                                        одного физического пикселя к размеру одного логического (CSS) пикселя.
                                        window.devicePixelRatio
                                        0 — выключает переопределение.
        :param mobile:              Эмуляция мобильного устройства. Это включает метатег области
                                        просмотра, полосы прокрутки наложения, авторазмер текста
                                        и многое другое.
        :param scale:               (optional, EXPERIMENTAL) Масштаб, применяемый к полученному изображению
                                        представления.
        :param screenWidth:         (optional, EXPERIMENTAL) Переопределяет значения ширины экрана в пикселях
                                        (от 0 до 10_000_000).
        :param screenHeight:        (optional, EXPERIMENTAL) Переопределяет значения высоты экрана в пикселях
                                        (от 0 до 10_000_000).
        :param positionX:           (optional, EXPERIMENTAL) Переопределяет X-позицию области просмотра на
                                        экране, в пикселях (от 0 до 10_000_000).
        :param positionY:           (optional, EXPERIMENTAL) Переопределяет Y-позицию области просмотра на
                                        экране, в пикселях (от 0 до 10_000_000).
        :param dontSetVisibleSize:  (optional, EXPERIMENTAL) Не устанавливайте видимый размер представления,
                                        полагайтесь на явный вызов setVisibleSize(смотри другие методы).
        :param screenOrientation:   (optional, EXPERIMENTAL) Переопределяет ориентацию экрана. Допустимые значения:
                                        {
                                            "type": str,    -> тип ориентации: portraitPrimary,
                                                                portraitSecondary, landscapePrimary,
                                                                landscapeSecondary
                                            "angle": int    -> Угол ориентации
                                        }
        :param viewport:            (optional, EXPERIMENTAL) Если установлено, видимая область страницы будет
                                        переопределена в этом окне просмотра. Это изменение окна просмотра
                                        не наблюдается на странице, например, относящиеся к области
                                        просмотра элементы не меняют позиции. Допустимые значения:
                                        {
                                            "x": float,     -> Смещение по оси X в независимых от устройства
                                                                пикселях (dip).
                                            "y": float,     -> Смещение по оси Y в независимых от устройства
                                                                пикселях (dip).
                                            "width": float, -> Ширина прямоугольника в независимых от устройства
                                                                пикселях (dip).
                                            "height": float,-> Высота прямоугольника в независимых от устройства
                                                                пикселях (dip).
                                            "scale": float  -> Коэффициент масштабирования страницы.
                                        }
        :return:
        """
        args = {"width": width, "height": height, "deviceScaleFactor": deviceScaleFactor, "mobile": mobile}
        if scale is not None:
            args.update({"scale": scale})
        if screenWidth is not None:
            args.update({"screenWidth": screenWidth})
        if screenHeight is not None:
            args.update({"screenHeight": screenHeight})
        if positionX is not None:
            args.update({"positionX": positionX})
        if positionY is not None:
            args.update({"positionY": positionY})
        if dontSetVisibleSize is not None:
            args.update({"dontSetVisibleSize": dontSetVisibleSize})
        if screenOrientation is not None:
            args.update({"screenOrientation": screenOrientation})
        if viewport is not None:
            args.update({"viewport": viewport})
        await self.Call("Emulation.setDeviceMetricsOverride", args)

    async def SetScrollbarsHidden(self, hidden: bool) -> None:
        """
        (EXPERIMENTAL)
        https://chromedevtools.github.io/devtools-protocol/tot/Emulation#method-setScrollbarsHidden
        :param hidden:              Должны ли полосы прокрутки быть всегда скрыты.
        :return:
        """
        await self.Call("Emulation.setScrollbarsHidden", {"hidden": hidden})

    async def SetDocumentCookieDisabled(self, disabled: bool) -> None:
        """
        (EXPERIMENTAL)
        https://chromedevtools.github.io/devtools-protocol/tot/Emulation#method-setDocumentCookieDisabled
        :param disabled:            Должен ли API document.cookie быть отключен.
        :return:
        """
        await self.Call("Emulation.setDocumentCookieDisabled", {"disabled": disabled})

    async def SetEmitTouchEventsForMouse(self, enabled: bool, configuration: Optional[str] = None) -> None:
        """
        (EXPERIMENTAL)
        https://chromedevtools.github.io/devtools-protocol/tot/Emulation#method-setEmitTouchEventsForMouse
        :param enabled:             Должна ли быть включена эмуляция касания на основе ввода от мыши.
        :param configuration:       (optional) Конфигурация событий касания / жеста.
                                        По умолчанию: текущая платформа. Допустимые значения:
                                            mobile, desktop
        :return:
        """
        args = {"enabled": enabled}
        if configuration is not None:
            args.update({"configuration": configuration})
        await self.Call("Emulation.setEmitTouchEventsForMouse", args)

    async def SetEmulatedMedia(self, media: str = "", features: Optional[list] = None) -> None:
        """
        Эмулирует переданный тип медиа или медиа-функцию для медиа-запросов CSS.
        https://chromedevtools.github.io/devtools-protocol/tot/Emulation#method-setEmulatedMedia
        :param media:               (optional) Тип медиа для эмуляции. Пустая строка отключает переопределение.
        :param features:            (optional) Список медиа-функций для эмуляции. Допустимые значения:
                                        [{
                                            "name": str,
                                            "value": str
                                        }, { ... }, ... ]
                                        https://chromedevtools.github.io/devtools-protocol/tot/Emulation#type-MediaFeature
        :return:
        """
        args = {"media": media}
        if features is not None:
            args.update({"features": features})
        await self.Call("Emulation.setEmulatedMedia", args)

    async def SetEmulatedVisionDeficiency(self, type_: str = "none") -> None:
        """
        (EXPERIMENTAL)
        Эмулирует переданный дефицит зрения.
        https://chromedevtools.github.io/devtools-protocol/tot/Emulation#method-setEmulatedVisionDeficiency
        :param type_:                (optional) Недостаток зрения для эмуляции. Допустимые значения:
                                        none, achromatopsia, blurredVision, deuteranopia, protanopia, tritanopia
        :return:
        """
        await self.Call("Emulation.setEmulatedVisionDeficiency", {"type": type_})

    async def SetGeolocationOverride(
            self,
             latitude: Optional[float] = None,
            longitude: Optional[float] = None,
             accuracy: Optional[float] = None
    ) -> None:
        """
        Переопределяет Положение или Ошибку Геолокации. Пропуск любого из параметров эмулирует положение недоступно.
        https://chromedevtools.github.io/devtools-protocol/tot/Emulation#method-setGeolocationOverride
        :param latitude:            (optional) Широта.
        :param longitude:           (optional) Долгота.
        :param accuracy:            (optional) Точность.
        :return:
        """
        args = {}
        if latitude is not None:
            args.update({"latitude": latitude})
        if longitude is not None:
            args.update({"longitude": longitude})
        if accuracy is not None:
            args.update({"accuracy": accuracy})
        await self.Call("Emulation.setGeolocationOverride", args)

    async def SetPageScaleFactor(self, pageScaleFactor: float) -> None:
        """
        (EXPERIMENTAL)
        Устанавливает переданный коэффициент масштабирования страницы.
        https://chromedevtools.github.io/devtools-protocol/tot/Emulation#method-setPageScaleFactor
        :param pageScaleFactor:     Коэффициент масштабирования страницы.
        :return:
        """
        await self.Call("Emulation.setPageScaleFactor", {"pageScaleFactor": pageScaleFactor})

    async def SetScriptExecutionDisabled(self, value: bool) -> None:
        """
        Переключает выполнение скриптов на странице.
        https://chromedevtools.github.io/devtools-protocol/tot/Emulation#method-setScriptExecutionDisabled
        :param value:               Должно ли выполнение скриптов быть отключено на странице.
        :return:
        """
        await self.Call("Emulation.setScriptExecutionDisabled", {"value": value})

    async def SetTouchEmulationEnabled(self, enabled: bool, maxTouchPoints: Optional[int] = None) -> None:
        """
        Включает "касания" для платформ, которые их не поддерживают.
        https://chromedevtools.github.io/devtools-protocol/tot/Emulation#method-setTouchEmulationEnabled
        :param enabled:             Должна ли эмуляция сенсорного события быть включена.
        :param maxTouchPoints:      (optional) Максимальное количество поддерживаемых точек касания.
                                        По умолчанию 1.
        :return:
        """
        args = {"enabled": enabled}
        if maxTouchPoints is not None:
            args.update({"maxTouchPoints": maxTouchPoints})
        await self.Call("Emulation.setTouchEmulationEnabled", args)

    async def SetLocaleOverride(self, locale: Optional[str] = None) -> None:
        """
        Не работает.
        https://bugs.chromium.org/p/chromium/issues/detail?id=1073363
        Починили с версии Хромиум == 83.0.4103.97

        (EXPERIMENTAL)
        Переопределяет языковой стандарт хост-системы на указанную.
        https://chromedevtools.github.io/devtools-protocol/tot/Emulation#method-setLocaleOverride
        :param locale:              (optional) Язык C в стиле ICU (например, "en_US"). Если не указано
                                        или не задано, отключает переопределение и восстанавливает
                                        локаль системы хоста по умолчанию.
                                        https://stackoverflow.com/questions/3191664/list-of-all-locales-and-their-short-codes
        :return:
        """
        args = {}
        if locale is not None:
            args.update({"locale": locale})
        await self.Call("Emulation.setLocaleOverride", args)

    async def SetTimezoneOverride(self, timezoneId: str = "") -> None:
        """
        (EXPERIMENTAL)
        Переопределяет часовой пояс хост-системы на указанный.
        https://chromedevtools.github.io/devtools-protocol/tot/Emulation#method-setTimezoneOverride
        :param timezoneId:          Идентификатор часового пояса("Europe/Moscow"). Если пусто,
                                        отключает переопределение и восстанавливает часовой пояс
                                        хост-системы по умолчанию.
                                        https://en.wikipedia.org/wiki/List_of_tz_database_time_zones
        :return:
        """
        await self.Call("Emulation.setTimezoneOverride", {"timezoneId": timezoneId})

    async def SetVisibleSize(self, width: int, height: int) -> None:
        """
        (DEPRECATED)
        (EXPERIMENTAL)
        Изменяет размер фрейма / области просмотра страницы. Обратите внимание, что это не влияет
            на контейнер фрейма (например, окно браузера). Может использоваться для создания
            скриншотов указанного размера. Не поддерживается на Android.
        https://chromedevtools.github.io/devtools-protocol/tot/Emulation#method-setVisibleSize
        :param width:               Ширина фрейма (DIP).
        :param height:              Высота фрейма (DIP).
        :return:            None
        """
        await self.Call("Emulation.setVisibleSize", {"width": width, "height": height})

    async def EmulationSetUserAgent(
            self, userAgent: str,
                acceptLanguage: Optional[str] = None,
                      platform: Optional[str] = None,
            userAgentMetadata: Optional[dict] = None
    ) -> None:
        """
        Переопределяет юзер-агент переданной строкой.
        https://chromedevtools.github.io/devtools-protocol/tot/Emulation#method-setUserAgentOverride
        :param userAgent:           Новый юзер-агент.
        :param acceptLanguage:      (optional) Язык браузера для эмуляции.
        :param platform:            (optional) Платформа браузера, которую возвращает
                                        "navigator.platform".
                                        https://www.w3schools.com/jsref/prop_nav_platform.asp
                                        https://stackoverflow.com/questions/19877924/what-is-the-list-of-possible-values-for-navigator-platform-as-of-today
        :param userAgentMetadata:   (optional, EXPERIMENTAL) Для отправки в заголовках Sec-CH-UA- * и возврата в
                                        navigator.userAgentData. Ожидатся словарь вида:
                                        {
                                            "brands": [{"brand": "brand name", "version": "brand version"}, { ... }, ... ],
                                            "fullVersion": "full version",
                                            "platform": "platform name",
                                            "platformVersion": "platform version",
                                            "architecture": "devise architecture",
                                            "model": "model",
                                            "mobile": boolean,
                                        }
        :return:            None
        """
        args = {"userAgent": userAgent}
        if acceptLanguage is not None:
            args.update({"acceptLanguage": acceptLanguage})
        if platform is not None:
            args.update({"platform": platform})
        if userAgentMetadata is not None:
            args.update({"userAgentMetadata": userAgentMetadata})
        await self.Call("Emulation.setUserAgentOverride", args)

    @abstractmethod
    async def Call(
            self, domain_and_method: str,
            params: Optional[dict] = None,
            wait_for_response: bool = True
    ) -> Union[dict, None]: raise NotImplementedError("async method Call() — is not implemented")

class EmulationEvent(DomainEvent):
    virtualTimeBudgetExpired = "Emulation.virtualTimeBudgetExpired"



from abc import ABC, abstractmethod
from typing import Optional, Union, List, Awaitable, Callable
from ..Data import DomainEvent
from dataclasses import dataclass

class Target(ABC):
    """
    #   https://chromedevtools.github.io/devtools-protocol/tot/Target
    """
    __slots__ = ()

    def __init__(self):
        self.targets_discovered = False

    @property
    def connected(self) -> bool:
        return False

    @property
    def verbose(self) -> bool:
        return False

    @property
    def is_headless_mode(self) -> bool:
        return False

    @property
    def page_id(self) -> str:
        return ""

    async def CreateBrowserContext(
        self,
        disposeOnDetach: Optional[bool] = None,
        proxyServer: Optional[str]      = None,
        proxyBypassList: Optional[str]  = None
    ) -> str:
        """
        Создает новый пустой BrowserContext. Аналогичен профилю в режиме инкогнито, но их может быть несколько.
        https://chromedevtools.github.io/devtools-protocol/tot/Target/#method-createBrowserContext
        :param disposeOnDetach:     Если True — удаляет контекст при отключении сеанса отладки.
        :param proxyServer:         Прокси-сервер, аналогичный тому, который передан --proxy-server.
        :param proxyBypassList:     Список обхода прокси, аналогичный тому, который передается в --proxy-bypass-list.
        :return:                Browser.BrowserContextID
        """
        args = {}
        if disposeOnDetach is not None: args.update(disposeOnDetach=disposeOnDetach)
        if proxyServer is not None: args.update(proxyServer=proxyServer)
        if proxyBypassList is not None: args.update(proxyBypassList=proxyBypassList)
        return (await self.Call("Target.createBrowserContext", args))["browserContextId"]

    async def GetBrowserContexts(self) -> List[str]:
        """
        Возвращает все контексты браузера, созданные с помощью метода Target.createBrowserContext.
        https://chromedevtools.github.io/devtools-protocol/tot/Target/#method-getBrowserContexts
        :return:                [ Browser.BrowserContextID, Browser.BrowserContextID, ... ]
        """
        return (await self.Call("Target.getBrowserContexts"))["browserContextIds"]

    async def DisposeBrowserContext(self, browserContextId: str) -> None:
        """
        Удаляет BrowserContext. Все соответствующие страницы будут закрыты без вызова их хуков beforeunload.
        https://chromedevtools.github.io/devtools-protocol/tot/Target/#method-disposeBrowserContext
        :param browserContextId:    Идентификатор контекста браузера.
        :return:
        """
        await self.Call("Target.disposeBrowserContext", {"browserContextId": browserContextId})

    async def GetTargetInfo(self, targetId: Optional[str] = None) -> 'TargetType.TargetInfo':
        """
        (EXPERIMENTAL)
        Возвращает информацию о "target", или о себе, если идентификатор не передан.
        https://chromedevtools.github.io/devtools-protocol/tot/Target#method-getTargetInfo
        :param targetId:            Идентификатор страницы/сервиса/воркера/...
        :return:                    targetInfo -> {
                                        "targetId":         str,
                                        "type":             str,
                                        "title":            str,
                                        "url":              str,
                                        "attached":         bool,
                                        "openerId":         str,
                                        "browserContextId": str,
                                    }
        """
        if targetId is None: targetId = self.page_id
        return TargetType.TargetInfo(**((await self.Call("Target.getTargetInfo", {"targetId": targetId}))["targetInfo"]))

    async def GetTargets(self) -> List['TargetType.TargetInfo']:
        """
        Возвращает список 'targetInfo' о доступных 'targets'.
        https://chromedevtools.github.io/devtools-protocol/tot/Target#method-getTargets
        :return:                [ targetInfo, targetInfo, ... ]
        """
        result = (await self.Call("Target.getTargets"))["targetInfos"]
        return [TargetType.TargetInfo(**info) for i, info in enumerate(result)]

    async def AttachToBrowserTarget(self) -> str:
        """
        Присоединяется к target браузера, использует только режим flat sessionId.
        https://chromedevtools.github.io/devtools-protocol/tot/Target#method-attachToBrowserTarget
        :return:                    sessionId
        """
        return (await self.Call("Target.attachToBrowserTarget"))["sessionId"]

    async def AttachToTarget(self, targetId: str, flatten: Optional[bool] = None) -> str:
        """
        Присоединяется к 'target' по указанному 'targetId'.
        https://chromedevtools.github.io/devtools-protocol/tot/Target#method-attachToTarget
        :param targetId:        Строка, представляющая идентификатор созданной страницы.
        :param flatten:         (optional) Разрешает "flat" доступ к сеансу с помощью указания атрибута
                                    sessionId в командах.
        :return:                sessionId -> Идентификатор, назначенный сеансу.
        """
        args = {"targetId": targetId}
        if flatten is not None: args.update({"flatten": flatten})
        return (await self.Call("Target.attachToTarget", args))["sessionId"]

    async def DetachFromTarget(self, sessionId: str = "", targetId: str = "") -> None:
        """
        Отключается от сессии переданного 'sessionId'.
        https://chromedevtools.github.io/devtools-protocol/tot/Target#method-detachFromTarget
        :param sessionId:       (optional) Строка, представляющая идентификатор сессии.
        :param targetId:        (optional) Строка, представляющая идентификатор созданной страницы.
        :return:
        """
        args = {}
        if sessionId: args.update({"sessionId": sessionId})
        elif targetId: args.update({"targetId": targetId})
        else: raise ValueError("At least one parameter must be specified 'sessionId' or 'targetId'")
        await self.Call("Target.detachFromTarget", args)

    async def SetAutoAttach(
            self, autoAttach: bool, waitForDebuggerOnStart: bool, flatten: Optional[bool] = None) -> None:
        """
        Определяет, следует ли автоматически присоединяться к новым target, которые считаются связанными
            с этой. При включении также присоединяется ко всем существующим связанным целям. При
            выключении автоматически отсоединяется от всех присоединенных в данный момент целей. Это
            также удаляет все цели, добавленные autoAttachRelated из списка целей, чтобы отслеживать
            создание связанных целей.
        https://chromedevtools.github.io/devtools-protocol/tot/Target#method-setAutoAttach
        :param autoAttach:                  Следует ли присоединяться к связанным targets.
        :param waitForDebuggerOnStart:      Приостанавливать ли новые targets при присоединении к ним.
                                                Используйте Runtime.runIfWaitingForDebugger для запуска
                                                приостановленных targets.
        :return:
        """
        args = {"autoAttach": autoAttach, "waitForDebuggerOnStart": waitForDebuggerOnStart}
        if flatten is not None: args.update(flatten=flatten)
        await self.Call("Target.setAutoAttach", args)

    async def CreateTarget(
        self,
        url: str                        = "about:blank",
        width: Optional[int]            = None,
        height: Optional[int]           = None,
        browserContextId: Optional[str] = None,
        enableBeginFrameControl: bool   = False,
        newWindow: bool                 = False,
        background: bool                = False
    ) -> str:
        """
        Создаёт новую страницу и возвращает её идентификатор. Чтобы затем управлять новой
            вкладкой, воспользуйтесь методом инстанса самого браузера GetPageByID(), или
            сразу методом CreatePage(), который проделывает все эти операции под капотом
            и возвращает готовый инстанс новой страницы.
        https://chromedevtools.github.io/devtools-protocol/tot/Target#method-createTarget
        :param url:                         Адрес будет открыт при создании.
        :param width:                       Ширина в Density-independent Pixels(DIP). Chrome Headless only!
        :param height:                      Высота в (DIP). Chrome Headless only!
        :param browserContextId:            Контекст браузера, которому предназначается создание target.
        :param enableBeginFrameControl:     Будет ли BeginFrames для этой цели контролироваться с помощью DevTools
                                                (только безголовый хром, пока не поддерживается в MacOS, по
                                                умолчанию false).
        :param newWindow:                   Если 'True' — страница будет открыта в новом окне.
        :param background:                  Если 'True' — страница будет открыта в фоне.
        :return:                targetId -> строка, представляющая идентификатор созданной страницы.
        """
        args = {
            "url": url, "enableBeginFrameControl": enableBeginFrameControl,
            "newWindow": False if self.is_headless_mode else newWindow,
            "background": False if self.is_headless_mode else background
        }
        if width is not None: args.update(width=width)
        if height is not None: args.update(height=height)
        if browserContextId is not None: args.update(browserContextId=browserContextId)
        return (await self.Call("Target.createTarget", args))["targetId"]

    async def CloseTarget(self, targetId: Optional[str] = None) -> None:
        """
        Закрывает вкладку указанного идентификатора, или завершает собственный инстанс,
            если идентификатор не передан.
        https://chromedevtools.github.io/devtools-protocol/tot/Target#method-closeTarget
        :param targetId:        Строка, представляющая идентификатор созданной страницы.
        :return:                None
        """
        if targetId is None: targetId = self.page_id
        await self.Call("Target.closeTarget", {"targetId": targetId})

    async def Close(self) -> None:
        """
        Закрывает вкладку.
        """
        await self.CloseTarget()

    async def SetDiscoverTargets(
        self, discover: bool,
        message:   Optional[Callable[[str, str], Awaitable[None]]] = None,
        created:   Optional[Callable[['TargetType.TargetInfo'], Awaitable[None]]] = None,
        crashed:   Optional[Callable[[str, str, int], Awaitable[None]]] = None,
        changed:   Optional[Callable[['TargetType.TargetInfo'], Awaitable[None]]] = None,
        destroyed: Optional[Callable[[str], Awaitable[None]]] = None
    ) -> None:
        """
        Управляет обнаружением доступных 'targets' уведомляя об их состоянии с помощью событий
            targetCreated / targetInfoChanged / targetDestroyed.
        https://chromedevtools.github.io/devtools-protocol/tot/Target#method-setDiscoverTargets
        :param discover:            'True' — включает эту надстройку, 'False' — выключает.
        :param message:             Корутина вызываемая для события 'Target.receivedMessageFromTarget'.
        :param created:             Корутина вызываемая для события 'Target.targetCreated'.
        :param crashed:             Корутина вызываемая для события 'Target.targetCrashed'.
        :param changed:             Корутина вызываемая для события 'Target.targetInfoChanged'.
        :param destroyed:           Корутина вызываемая для события 'Target.targetDestroyed'.
        :return:
        """
        async def message_decor(params: dict, func: Callable[[str, str], Awaitable[None]]) -> None:
            await func(params["sessionId"], params["message"])

        async def crash_decor(params: dict, func: Callable[[str, str, int], Awaitable[None]]) -> None:
            await func(params["targetId"], params["status"], params["errorCode"])

        async def decorator(params: dict, func: Callable[['TargetType.TargetInfo'], Awaitable[None]]) -> None:
            await func(TargetType.TargetInfo(**params["targetInfo"]))

        async def destroy_decor(params: dict, func: Callable[[str], Awaitable[None]]) -> None:
            await func(params["targetId"])

        if discover:
            if message is not None: await self.AddListenerForEvent(TargetEvent.receivedMessageFromTarget, message_decor, message)
            if created is not None: await self.AddListenerForEvent(TargetEvent.targetCreated, decorator, created)
            if crashed is not None: await self.AddListenerForEvent(TargetEvent.targetCrashed, crash_decor, crashed)
            if changed is not None: await self.AddListenerForEvent(TargetEvent.targetInfoChanged, decorator, changed)
            if destroyed is not None: await self.AddListenerForEvent(TargetEvent.targetDestroyed, destroy_decor, destroyed)
        else:
            for event in [
                'Target.receivedMessageFromTarget',
                'Target.targetCreated',
                'Target.targetCrashed',
                'Target.targetInfoChanged',
                'Target.targetDestroyed'
            ]:
                self.RemoveListenersForEvent(event)
        self.targets_discovered = discover
        await self.Call("Target.setDiscoverTargets", {"discover": discover})

    @abstractmethod
    async def Call(
        self, domain_and_method: str,
        params: Optional[dict]            = None,
        wait_for_response: bool = True
    ) -> Union[dict, None]: raise NotImplementedError("async method Call() — is not implemented")

    @abstractmethod
    async def AddListenerForEvent(
            self, event: Union[str, DomainEvent], listener: Callable, *args: any) -> None:
        raise NotImplementedError("async method AddListenerForEvent() — is not implemented")

    @abstractmethod
    def RemoveListenersForEvent(self, event: str) -> None:
        raise NotImplementedError("method RemoveListenersForEvent() — is not implemented")


class TargetEvent(DomainEvent):
    receivedMessageFromTarget = "Target.receivedMessageFromTarget"
    targetCrashed = "Target.targetCrashed"
    targetCreated = "Target.targetCreated"
    targetDestroyed = "Target.targetDestroyed"
    targetInfoChanged = "Target.targetInfoChanged"
    attachedToTarget = "Target.attachedToTarget"                        # ! EXPERIMENTAL
    detachedFromTarget = "Target.detachedFromTarget"                    # ! EXPERIMENTAL


class TargetType:

    @dataclass
    class RemoteLocation:
        host: str; port: int


    @dataclass
    class TargetInfo:
        targetId: str; type: str; title: str; url: str; attached: bool
        openerId:         Optional[str] = None
        canAccessOpener: Optional[bool] = None
        openerFrameId:    Optional[str] = None
        browserContextId: Optional[str] = None



import re
from abc import ABC, abstractmethod
from typing import Optional, Union, List
from ..DOMElement import Node
from ..domains.Runtime import RuntimeType
from ..exceptions import CouldNotFindNodeWithGivenID, RootIDNoLongerExists
from ..Data import DomainEvent

class DOM(ABC):
    """
    #   https://chromedevtools.github.io/devtools-protocol/tot/DOM
    """
    __slots__ = ()

    def __init__(self):
        self.dom_domain_enabled = False

    @property
    def connected(self) -> bool:
        return False

    @property
    def verbose(self) -> bool:
        return False

    @property
    def page_id(self) -> str:
        return ""

    async def DOMEnable(self) -> None:
        """
        Включает DOM-агент для данной страницы.
        https://chromedevtools.github.io/devtools-protocol/tot/DOM#method-enable
        :return:
        """
        await self.Call("DOM.enable")
        self.dom_domain_enabled = True

    async def DOMDisable(self) -> None:
        """
        Отключает DOM-агент для данной страницы.
        https://chromedevtools.github.io/devtools-protocol/tot/DOM#method-disable
        :return:
        """
        await self.Call("DOM.disable")
        self.dom_domain_enabled = False

    async def GetRoot(self, depth: Optional[int] = None, pierce: Optional[bool] = None) -> Node:
        """
        Возвращает корневой узел документа.
        https://chromedevtools.github.io/devtools-protocol/tot/DOM#method-getDocument
        Корневой элемент ВСЕГДА имеет подобную структуру:
        'root': {
            'nodeId': 19,
            'backendNodeId': 2,
            'nodeType': 9,
            'nodeName': '#document',
            'localName': '',
            'nodeValue': '',
            'childNodeCount': 2,
            'children': [
                {
                    'nodeId': 20,
                    'parentId': 19,
                    'backendNodeId': 9,
                    'nodeType': 10,
                    'nodeName': 'html',
                    'localName': '',
                    'nodeValue': '',
                    'publicId': '',
                    'systemId': ''
                }, {
                    'nodeId': 21,
                    'parentId': 19,
                    'backendNodeId': 10,
                    'nodeType': 1,
                    'nodeName': 'HTML',
                    'localName': 'html',
                    'nodeValue': '',
                    'childNodeCount': 2,
                    'children': [
                        {
                            'nodeId': 22,
                            'parentId': 21,
                            'backendNodeId': 11,
                            'nodeType': 1,
                            'nodeName': 'HEAD',
                            'localName': 'head',
                            'nodeValue': '',
                            'childNodeCount': 4,
                            'attributes': [ ]
                        }, {
                            'nodeId': 23,
                            'parentId': 21,
                            'backendNodeId': 12,
                            'nodeType': 1,
                            'nodeName': 'BODY',
                            'localName': 'body',
                            'nodeValue': '',
                            'childNodeCount': 8,
                            'attributes': [ ]
                        }
                    ],
                    'attributes': [
                        'lang',
                        'ru'
                    ],
                    'frameId': 'AF11E1D7BC9DF951D77C6C07C02B98E7'
                }
            ],
            'documentURL': 'url ...',
            'baseURL': 'url ...',
            'xmlVersion': ''
        }
        :param depth:           Максимальная глубина, на которой должны быть извлечены
                                    дочерние элементы, по умолчанию равна 1. Используйте
                                    -1 для всего поддерева или укажите целое число больше 0.
        :param pierce:          Должны ли проходиться iframes и теневые корни при возврате
                                    поддерева (по умолчанию false).
        :return:            <Node>.
        """
        args = {}
        if depth is not None: args.update(depth=depth)
        if pierce is not None: args.update(pierce=pierce)
        node: dict = (await self.Call("DOM.getDocument", args))["root"]
        return Node(self, **node)

    async def QuerySelector(
            self, selector: str,
            ignore_root_id_exists: bool = False,
            in_frames: bool = False
    ) -> Union[Node, None]:
        """
        Выполняет DOM-запрос, возвращая объект найденного узла, или None.
            Эквивалент  === document.querySelector()
        https://chromedevtools.github.io/devtools-protocol/tot/DOM#method-querySelector
        :param selector:                    Селектор.
        :param ignore_root_id_exists:       Игнорировать исключение при отсутствии родительского элемента.
                                                Полезно при запросах на загружающихся страницах.
        :param in_frames:                   Опрашивать документ вкючая shadow-root и iframe?
        :return:                <Node>
        """
        args = {} if not in_frames else dict(depth=-1, pierce=True)
        root_node_id = (await self.Call("DOM.getDocument", args))["root"]["nodeId"]
        try:
            node: dict = await self.Call("DOM.querySelector", {
                "nodeId": root_node_id, "selector": selector
            })
        except CouldNotFindNodeWithGivenID as e:
            if match := re.search(r"nodeId\': (\d+)", str(e)):
                if match.group(1) == str(root_node_id):
                    if ignore_root_id_exists:
                        return None
                    raise RootIDNoLongerExists
            raise
        return Node(self, **node) if node["nodeId"] > 0 else None


    async def QuerySelectorAll(
            self, selector: str,
            ignore_root_id_exists: bool = False,
            in_frames: bool = False
    ) -> List[Node]:
        """
        Выполняет DOM-запрос, возвращая список объектов найденных узлов, или пустой список.
            Эквивалент  === document.querySelectorAll()
        https://chromedevtools.github.io/devtools-protocol/tot/DOM#method-querySelectorAll
        :param selector:                    Селектор.
        :param ignore_root_id_exists:       Игнорировать исключение при отсутствии родительского элемента.
                                                Полезно при запросах на загружающихся страницах.
        :param in_frames:                   Опрашивать документ вкючая shadow-root и iframe?
        :return:                [ <Node>, <Node>, ... ]
        """
        nodes = []
        args = {} if not in_frames else dict(depth=-1, pierce=True)
        root_node_id = (await self.Call("DOM.getDocument", args))["root"]["nodeId"]
        try:
            for node in (await self.Call("DOM.querySelectorAll", {
                "nodeId": root_node_id, "selector": selector
            }))["nodeIds"]:
                nodes.append(Node(self, node))
        except CouldNotFindNodeWithGivenID as e:
            if match := re.search(r"nodeId\': (\d+)", str(e)):
                if match.group(1) == str(root_node_id):
                    if ignore_root_id_exists:
                        return []
                    raise RootIDNoLongerExists
            raise
        return nodes

    async def PerformSearch(self, query: str, searchInShadowDOM: Optional[bool] = None) -> dict:
        """
        (EXPERIMENTAL)
        Ищет заданную строку в дереве DOM. Используйте 'GetSearchResults()' для доступа к результатам
            поиска или 'CancelSearch()'( !не найдено! ), чтобы завершить этот сеанс поиска. DOM-агент
            должен быть влючён.
        https://chromedevtools.github.io/devtools-protocol/tot/DOM#method-performSearch
        :param query:               Обычный текст, селектор, или поисковый запрос XPath.
        :param searchInShadowDOM:   (optional) True - поиск будет так же выполнен в shadow DOM.
        :return:                    {"searchId": str(searchId), "resultCount": int(resultCount)}
                                        searchId    - уникальный идентификатор сессии поиска.
                                        resultCount - кол-во результатов удовлетворяющих запрос.
        """
        args = {"query": query}
        if searchInShadowDOM is not None:
            args.update({"includeUserAgentShadowDOM": searchInShadowDOM})
        return await self.Call("DOM.performSearch", args)

    async def GetSearchResults(
            self, searchId: str,
            fromIndex: int = 0,
            toIndex:   int = 0
    ) -> List[Node]:
        """
        (EXPERIMENTAL)
        Возвращает список результатов поиска для поисковой сессии 'searchId', в интервале от 'fromIndex'
            до 'toIndex', полученной в результате вызова PerformSearch().
        https://chromedevtools.github.io/devtools-protocol/tot/DOM#method-getSearchResults
        :param searchId:        Уникальный идентификатор сессии поиска.
        :param fromIndex:       Начальный индекс результата поиска, который будет возвращен.
        :param toIndex:         Конечный индекс результата поиска, который будет возвращен.
        :return:                [ <Node>, <Node>, ... ]
        """
        nodes = []
        args = {"searchId": searchId, "fromIndex": fromIndex, "toIndex": toIndex}
        for node_id in (await self.Call("DOM.getSearchResults", args))["nodeIds"]:
            if self.verbose:
                print("[SearchResults] node_id =", node_id)
            nodes.append(Node(self, node_id, ""))
        return nodes

    async def Undo(self) -> None:
        """
        (EXPERIMENTAL)
        Отменяет последнее выполненное действие.
        https://chromedevtools.github.io/devtools-protocol/tot/DOM#method-undo
        :return:
        """
        await self.Call("DOM.undo")

    async def Redo(self) -> None:
        """
        (EXPERIMENTAL)
        Повторно выполняет последнее отмененное действие.
        https://chromedevtools.github.io/devtools-protocol/tot/DOM/#method-redo
        :return:
        """
        await self.Call("DOM.redo")

    async def MarkUndoableState(self) -> None:
        """
        (EXPERIMENTAL)
        Отмечает последнее состояние, которое нельзя изменить.
        https://chromedevtools.github.io/devtools-protocol/tot/DOM/#method-markUndoableState
        :return:
        """
        await self.Call("DOM.markUndoableState")

    async def DescribeNode(
            self, nodeId: Optional[int] = None,
            backendNodeId: Optional[int] = None,
            objectId: Optional[str] = None,
            depth: Optional[int] = None,
            pierce: Optional[bool] = None
    ) -> Node:
        """
        Описывает узел с учетом его идентификатора, не требует включения домена. Не начинает отслеживать какие-либо
            объекты, можно использовать для автоматизации.
        https://chromedevtools.github.io/devtools-protocol/tot/DOM/#method-describeNode
        :return:
        """
        if not ((nodeId != None) | (backendNodeId != None) | (objectId != None)):
            raise ValueError("Один из nodeId, backendNodeId, или objectId — должен присутствовать!")
        args = {}
        if nodeId is not None: args.update(nodeId=nodeId)
        if backendNodeId is not None: args.update(backendNodeId=backendNodeId)
        if objectId is not None: args.update(objectId=objectId)
        if depth is not None: args.update(depth=depth)
        if pierce is not None: args.update(pierce=pierce)
        result = await self.Call("DOM.describeNode", args)
        return Node(self, **result["node"])

    async def ResolveNode(
            self, nodeId: Optional[int] = None,
            backendNodeId: Optional[int] = None,
            objectGroup: Optional[str] = None,
            executionContextId: Optional[str] = None
    ) -> RuntimeType.RemoteObject:
        """
        Создаёт JavaScript-объект для указанной ноды и возвращает его описание.
        https://chromedevtools.github.io/devtools-protocol/tot/DOM/#method-resolveNode
        :return:
        """
        if not ((nodeId != None) | (backendNodeId != None)):
            raise ValueError("Один из nodeId, или backendNodeId — должен присутствовать!")
        args = {}
        if nodeId is not None: args.update(nodeId=nodeId)
        if backendNodeId is not None: args.update(backendNodeId=backendNodeId)
        if objectGroup is not None: args.update(objectGroup=objectGroup)
        if executionContextId is not None: args.update(executionContextId=executionContextId)
        result: dict = await self.Call("DOM.resolveNode", args)
        return RuntimeType.RemoteObject(**result.get("object"))

    async def RequestNode(self, objectId: str) -> Node:
        """
        Запрашивает, чтобы узел был отправлен вызывающей стороне с учетом ссылки на объект узла JavaScript.
            Все узлы, формирующие путь от узла к корню, также отправляются клиенту в виде серии
            setChildNodes-уведомлений.
        https://chromedevtools.github.io/devtools-protocol/tot/DOM/#method-requestNode
        :return:
        """
        args = {"objectId": objectId}
        result: dict = await self.Call("DOM.requestNode", args)
        return Node(self, result.get("nodeId"))

    @abstractmethod
    async def Call(
            self, domain_and_method: str,
            params: Optional[dict] = None,
            wait_for_response: bool = True
    ) -> Union[dict, None]: raise NotImplementedError("async method Call() — is not implemented")


class DOMEvent(DomainEvent):
    attributeModified = "DOM.attributeModified"
    attributeRemoved = "DOM.attributeRemoved"
    characterDataModified = "DOM.characterDataModified"
    childNodeCountUpdated = "DOM.childNodeCountUpdated"
    childNodeInserted = "DOM.childNodeInserted"
    childNodeRemoved = "DOM.childNodeRemoved"
    documentUpdated = "DOM.documentUpdated"
    setChildNodes = "DOM.setChildNodes"
    distributedNodesUpdated = "DOM.distributedNodesUpdated"     # ! EXPERIMENTAL
    inlineStyleInvalidated = "DOM.inlineStyleInvalidated"       # ! EXPERIMENTAL
    pseudoElementAdded = "DOM.pseudoElementAdded"               # ! EXPERIMENTAL
    pseudoElementRemoved = "DOM.pseudoElementRemoved"           # ! EXPERIMENTAL
    shadowRootPopped = "DOM.shadowRootPopped"                   # ! EXPERIMENTAL
    shadowRootPushed = "DOM.shadowRootPushed"                   # ! EXPERIMENTAL



from typing import Optional, Type
import re


class MyBaseException(Exception): pass

class JavaScriptError(MyBaseException): pass

class NullProperty(MyBaseException): pass

class EvaluateError(MyBaseException): pass

class PromiseEvaluateError(MyBaseException): pass

class NodeNotResolved(MyBaseException): pass

class NodeNotDescribed(MyBaseException): pass

class StateError(MyBaseException): pass

class FlagArgumentContainError(MyBaseException): pass


class TargetCrashed(MyBaseException): pass

class PositionOutOfBounds(MyBaseException): pass

class CouldNotFindNodeWithGivenID(MyBaseException): pass

class RootIDNoLongerExists(MyBaseException): pass                       # ! для QuerySelector-методов

class CouldNotComputeContentQuads(MyBaseException): pass

class NoDialogIsShowing(MyBaseException): pass                          # ! при перехвате диалоговых окон

class NoTargetWithGivenIdFound(MyBaseException): pass

class NoScriptWithGivenId(MyBaseException): pass

class UniqueContextIdNotFound(MyBaseException): pass

class AnotherLocaleOverrideIsAlreadyInEffect(MyBaseException): pass     # ! при установке той же локали

class FontFamiliesCanOnlyBeSetOnce(MyBaseException): pass               # ! при установке тех же шрифтов


exception_store = {
    "Target crashed": TargetCrashed,
    "Position out of bounds": PositionOutOfBounds,
    "Could not find node with given id": CouldNotFindNodeWithGivenID,
    "Could not compute content quads": CouldNotComputeContentQuads,
    "No dialog is showing": NoDialogIsShowing,
    "No target with given id found": NoTargetWithGivenIdFound,
    "No script with given id": NoScriptWithGivenId,
    "uniqueContextId not found": UniqueContextIdNotFound,
    "Another locale override is already in effect": AnotherLocaleOverrideIsAlreadyInEffect,
    "Font families can only be set once": FontFamiliesCanOnlyBeSetOnce
}


def get_cdtp_error(error_text: str) -> Optional[Type[MyBaseException]]:
    for title, ex in exception_store.items():
        if title in error_text:
            return ex
    return None
    

def highlight_eval_error(error_text: str, expression: str) -> str:
    """ Подсвечивает место в JS-коде ставшее причиной исключения.
    """
    line, pos = tuple(map(int, error_text.split(":")[-2:]))
    lines = expression.split("\n")
    l = lines[line - 1]
    word = re.match(r"\w+\b", l[pos - 1:]).group(0)
    l = "".join([l[:pos - 1], f"\x1b[91m\x1b[4m{word}\x1b[0m", l[len(word) + pos - 1:]])
    return "\n".join([
        error_text,
        "\n\x1b[36mProblem in code is here: \x1b[37m",
        *lines[:line - 1], l, *lines[line:],
    ])


def highlight_promise_error(error_text: str) -> str:
    lines = error_text.split("\n")
    e_type, e_title = lines[0].split(":")
    print(error_text)
    return "".join([
        f"\x1b[36m{e_type}: \x1b[37m",
        f"\x1b[91m\x1b[4m{e_title.strip()}\x1b[0m",
        "\n", *lines[1:]
    ])



from abc import ABC, abstractmethod
from typing import Optional, Union, List
from ..Data import SystemData, GPUInfo, ProcessInfo

class SystemInfo(ABC):
    """
    #   https://chromedevtools.github.io/devtools-protocol/tot/SystemInfo/
    """
    __slots__ = ()

    @property
    def connected(self) -> bool:
        return False

    @property
    def verbose(self) -> bool:
        return False

    @property
    def page_id(self) -> str:
        return ""

    async def GetSystemInfo(self) -> SystemData:
        """
        Возвращает информацию о системе.
        https://chromedevtools.github.io/devtools-protocol/tot/SystemInfo/#method-getInfo
        :return:
        """
        result = await self.Call("SystemInfo.getInfo")
        return SystemData(
            gpu=GPUInfo(**result["gpu"]), modelName=result["modelName"],
            modelVersion=result["modelVersion"], commandLine=result["commandLine"]
        )

    async def GetProcessInfo (self) -> List[ProcessInfo]:
        """
        Возвращает информацию обо всех запущенных в системе процессах.
        https://chromedevtools.github.io/devtools-protocol/tot/SystemInfo/#method-getProcessInfo
        :return:
        """
        result = await self.Call("SystemInfo.getProcessInfo")
        return [ProcessInfo(**i) for i in result]

    @abstractmethod
    async def Call(
            self, domain_and_method: str,
            params: Optional[dict] = None,
            wait_for_response: bool = True
    ) -> Union[dict, None]: raise NotImplementedError("async method Call() — is not implemented")




__version__ = "0.9.3"
__author__ = "PieceOfGood"
__email__ = "78sanchezz@gmail.com"

__all__ = [
    "find_instances",
    "CMDFlags",
    "BrowserEx",
    "PageEx",
    "catch_headers_for_url"
]

from .Browser import CMDFlags
from .Browser import Browser
from .BrowserEx import BrowserEx
from .PageEx import PageEx, catch_headers_for_url

find_instances = Browser.FindInstances
